
# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import logging


try:
    from .version import genetest_version as __version__
except ImportError:
    __version__ = None


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__credits__ = ["Louis-Philippe Lemieux Perreault", "Marc-Andre Legault"]
__license__ = "CC BY-NC 4.0"
__status__ = "Development"


# The log format
_LOG_FORMAT = logging.Formatter(
    fmt="[%(asctime)s %(levelname)s] %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


def test(verbosity=1):
    """Execute all the tests."""
    import unittest
    from .tests import test_suite

    unittest.TextTestRunner(verbosity=verbosity).run(test_suite)



"""
Run a full statistical analysis.
"""


# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import time
import queue
import itertools
import multiprocessing
import traceback
import logging
from itertools import chain
import warnings

import numpy as np
import pandas as pd

from .statistics.core import StatsError
from .statistics.descriptive import get_maf
from . import subscribers as subscribers_module
from .modelspec import SNPs, PheWAS, modelspec_from_formula


logger = logging.getLogger(__name__)


def _missing(y, X):
    """Return a boolean vector indexing non-missing values."""
    y_missing = y.isnull().any(axis=1)
    X_missing = X.isnull().any(axis=1)
    return ~(y_missing | X_missing)


def _generate_sample_order(x_samples, geno_samples):
    """Create the order for samples between X and genotypes."""
    # Generating the order for the genotypes
    geno_order = np.array(
        [i for i, s in enumerate(geno_samples) if s in x_samples],
        dtype=int,
    )

    # Generating the order for the X matrix
    x_order = geno_samples.values[geno_order]

    if x_samples.duplicated().any():
        # The are duplicated samples, so we need to duplicate the genotype
        # indexes
        logger.debug("Duplicated samples found")
        counts = x_samples.value_counts()
        geno_order = np.array(
            list(chain(*[[geno_samples.get_loc(s)]*counts[s]
                         for s in x_order])),
            dtype=int,
        )

    return geno_order, x_order


def _gwas_worker(q, results_q, failed, abort, fit, y, X, samples, maf_t=None):
    # The sample order (to add genotypes to the X data frame
    geno_index = None
    sample_order = None

    # Get a SNP.
    while not abort.is_set():
        # Get a SNP from the Queue.
        try:
            snp = q.get(timeout=1)
        except queue.Empty:
            # We waited for 1 seconds, just in case abort was set and no more
            # SNP is getting pushed in the Queue
            continue

        # This is a check for a sentinel.
        if snp is None:
            break

        # Compute union between indices if not already done
        if sample_order is None or geno_index is None:
            # Checking the intersection
            geno_index, sample_order = _generate_sample_order(X.index, samples)

            # Do we have an intersect
            if geno_index.shape[0] == 0:
                logger.critical("Genotype and phenotype data have "
                                "non-overlapping indices.")
                abort.set()
                continue

        # Set all to missing genotypes
        X.loc[:, "SNPs"] = np.nan

        # Set the genotypes
        X.loc[sample_order, "SNPs"] = snp.genotypes[geno_index]

        not_missing = _missing(y, X)

        # Computing MAF
        maf, minor, major, flip = get_maf(
            genotypes=X.loc[not_missing, "SNPs"],
            minor=snp.coded,
            major=snp.reference,
        )

        # Is the MAF below the threshold?
        if maf_t is not None and maf < maf_t:
            failed.put((snp.variant.name, "MAF: {} < {}".format(maf, maf_t)))
            continue

        # Flipping if required
        if flip:
            X.loc[:, "SNPs"] = 2 - X.loc[:, "SNPs"]

        # Computing
        results = None
        try:
            with warnings.catch_warnings(record=True) as warning_list:
                results = fit(y[not_missing], X[not_missing])

                # Logging warnings
                _log_warnings(snp.variant.name, warning_list)

        except StatsError as e:
            logger.warning("{}: {}".format(snp.variant.name, e))
            if snp.variant.name:
                failed.put((snp.variant.name, str(e)))
            continue

        except Exception as e:
            logger.critical("{} was raised in worker\n{}".format(
                type(e).__name__, traceback.format_exc(),
            ))
            abort.set()
            continue

        # Update the results for the SNP with metadata.
        results["SNPs"].update({
            "chrom": snp.variant.chrom, "pos": snp.variant.pos, "major": major,
            "minor": minor, "name": snp.variant.name,
        })
        results["SNPs"]["maf"] = maf

        results_q.put(results)

    # The main loop was exited either because of an abort or because the
    # sentinel was encountered.
    # We put None to the queues to signal the worker is done and exit.
    q.put(None)
    results_q.put(None)
    failed.put(None)
    logger.debug("Worker Done")


def _log_warnings(identifier, warning_list):
    """Logs the warnings."""
    done = set()
    for w in [str(_.message) for _ in warning_list]:
        if w not in done:
            logger.warning("{}: {}".format(identifier, w))
            done.add(w)


def execute_formula(phenotypes, genotypes, formula, test, test_kwargs=None,
                    subscribers=None, variant_predicates=None,
                    output_prefix=None, maf_t=None, cpus=None):

    # Getting the model specification and the subgroups (if any)
    modelspec, subgroups = modelspec_from_formula(formula, test, test_kwargs)

    # Executing
    execute(phenotypes, genotypes, modelspec, subscribers, variant_predicates,
            output_prefix, subgroups, maf_t, cpus)


def execute(phenotypes, genotypes, modelspec, subscribers=None,
            variant_predicates=None, output_prefix=None, subgroups=None,
            maf_t=None, cpus=None):
    """Execute an analysis.

    Args:
        phenotypes (): The phenotypes container.
        genotypes (): The genotypes container.
        modelspec (genetest.modelspec.ModelSpec): The model specification.
        subscribers (list): A list of subscribers.
        variant_predicates (list): A list of variant predicates.
        output_prefix (str): The output prefix.
        subgroups (list): The subgroup analysis.
        maf_t (float): The MAF threshold.
        cpus (int): The number of CPUs to perform the analysis.

    """
    if subscribers is None:
        subscribers = [subscribers_module.Print()]

    # Initialize the subscribers.
    for subscriber in subscribers:
        subscriber.init(modelspec)

    if variant_predicates is None:
        variant_predicates = []

    # We branch out early if it is a pheWAS analysis because the data
    # preparation steps are pretty different.
    if isinstance(modelspec.outcome, PheWAS):
        return _execute_phewas(phenotypes, genotypes, modelspec, subscribers,
                               variant_predicates, output_prefix, subgroups,
                               cpus)

    data = modelspec.create_data_matrix(phenotypes, genotypes)

    # Exclude samples with missing outcome or covariable.
    data = data.dropna()

    # Extract y and X matrices
    y_cols = tuple(modelspec.outcome.keys())
    y = data[[modelspec.outcome[col].id for col in y_cols]]
    X = data.drop(y.columns, axis=1)

    # Rename y columns
    y.columns = y_cols

    # Drop uninformative factors.
    bad_cols = _get_uninformative_factors(X, modelspec)
    if len(bad_cols):
        logger.info(
            "After removing missing values, dropping ({}) factor levels that "
            "have no variation."
            "".format(len(bad_cols))
        )
        X = X.drop(bad_cols, axis=1)

    messages = {
        "skipped": [],
        "failed": []
    }

    if modelspec.stratify_by is not None:
        _execute_stratified(genotypes, modelspec, subscribers, y, X,
                            variant_predicates, subgroups, messages, maf_t,
                            cpus)
    elif SNPs in modelspec.predictors:
        _execute_gwas(genotypes, modelspec, subscribers, y, X,
                      variant_predicates, messages, maf_t, cpus)
    else:
        _execute_simple(modelspec, subscribers, y, X, variant_predicates,
                        messages)

    for subscriber in subscribers:
        subscriber.close()

    prefix = (output_prefix + "_") if output_prefix is not None else ""
    if messages["failed"]:
        with open(prefix + "failed_snps.txt", "w") as f:
            for failed_snp in messages["failed"]:
                f.write("{}\t{}\n".format(*failed_snp))

    if messages["skipped"]:
        with open(prefix + "not_analyzed_snps.txt", "w") as f:
            for failed_snp in messages["skipped"]:
                f.write("{}\n".format(failed_snp))


def _execute_phewas(phenotypes, genotypes, modelspec, subscribers,
                    variant_predicates, output_prefix, subgroups, cpus=None):
    # Check that invalid options were not set.
    if subgroups:
        raise NotImplementedError(
            "Stratified analyses for pheWAS are not supported yet."
        )

    if variant_predicates:
        raise ValueError(
            "Variant predicates can only be used in the context of GWAS "
            "analyses."
        )

    # Create the matrices.
    data = modelspec.create_data_matrix(phenotypes, genotypes)

    # Prepare synchronized data structures.
    results_queue = multiprocessing.Queue()
    phen_queue = multiprocessing.Queue()
    abort = multiprocessing.Event()

    # The number if CPUs
    if cpus is None:
        cpus = multiprocessing.cpu_count() - 1

    # Create workers.
    predictors = modelspec.predictors
    workers = []
    fit = modelspec.test().fit
    for worker in range(cpus):
        this_data = data.copy()

        worker = multiprocessing.Process(
            target=_phewas_worker,
            args=(this_data, predictors, abort, fit, phen_queue, results_queue)
        )

        workers.append(worker)
        worker.start()

    # Fill the phenotypes queue.
    for entity in modelspec.outcome.li:
        if entity not in modelspec.predictors:
            phen_queue.put(entity)

    phen_queue.put(None)

    # Pass results to the subscribers until it's done.
    translations = modelspec.get_translations()
    n_workers_done = 0
    while n_workers_done != len(workers):
        try:
            res = results_queue.get(False)
        except queue.Empty:
            time.sleep(1)
            continue

        if res is None:
            n_workers_done += 1
            continue

        res["outcome"] = translations[res["outcome"].id]
        for subscriber in subscribers:
            try:
                subscriber.handle(res)
            except KeyError as e:
                subscribers_module.subscriber_error(e.args[0], abort)

    for worker in workers:
        worker.join()

    logger.info("PheWAS complete.")


def _phewas_worker(data, predictors, abort, fit, phen_queue, results_queue):

    predictors = [i.id for i in predictors]
    if "intercept" in data.columns:
        predictors.append("intercept")

    X = data[predictors]

    while not abort.is_set():
        # Get an outcome.
        y = phen_queue.get()

        # Sentinel check.
        if y is None:
            phen_queue.put(None)
            results_queue.put(None)
            logger.debug("Worker Done")
            return

        y_data = data[[y.id]]
        not_missing = _missing(y_data, X)

        try:
            results = fit(y_data[not_missing], X[not_missing])
        except Exception as e:
            logger.debug("Exception raised during fitting:", e)
            print(e)
            continue

        results["outcome"] = y
        results_queue.put(results)


def _execute_stratified(genotypes, modelspec, subscribers, y, X,
                        variant_predicates, subgroups, messages, maf_t=None,
                        cpus=None):
    # Levels.
    assert len(modelspec.stratify_by) == len(subgroups)

    var_levels = []
    for i, subgroup in enumerate(subgroups):
        if subgroup is None:
            levels = X[modelspec.stratify_by[i].id].dropna().unique()
            var_levels.append(levels)
        elif hasattr(subgroup, "__iter__"):
            var_levels.append(subgroup)
        else:
            var_levels.append([subgroup])

    # This create a iterable of all level combinations to analyze:
    # assume x = 1, 2 and y = 3
    # Then subsets = [[1, 3], [2, 3]]
    # The order from modelspec.stratify_by is kept.
    subsets = itertools.product(*var_levels)

    gwas_mode = SNPs in modelspec.predictors
    translations = modelspec.get_translations()

    for levels in subsets:
        # current_subset is an iterable of (entity, level) pairs.
        current_subset = list(zip(modelspec.stratify_by, levels))

        # Build the filtering vector.
        idx = np.logical_and.reduce(
            [X[var.id] == level for var, level in current_subset]
        )

        # Extract the stratification and execute the analysis.
        subset_info = {
            translations[var.id]: level for var, level in current_subset
        }
        for sub in subscribers:
            sub._update_current_subset(subset_info)

        # Drop columns that become uninformative after stratification.
        this_x = X.loc[idx, :]
        bad_cols = _get_uninformative_factors(this_x, modelspec)
        this_x = this_x.drop(bad_cols, axis=1)

        # Also drop the columns from the stratification variables.
        this_x = this_x.drop([i.id for i in modelspec.stratify_by], axis=1)

        logger.info("Analysing subgroup {}".format(
            ";".join("{}:{}".format(*_) for _ in sorted(subset_info.items()))
        ))

        # Make sure everything went ok.
        if this_x.shape[0] == 0:
            raise ValueError(
                "No samples left in subgroup analysis ({}). Are all requested "
                "levels valid?"
                "".format(subset_info)
            )
        elif this_x.shape[1] == 0:
            raise ValueError(
                "No columns left in subgroup analysis ({}). Maybe there are "
                "no samples with non-null values in the requested subgroup."
                "".format(subset_info)
            )

        if len(bad_cols):
            logger.info(
                "After stratification, dropping factor levels ({}) that have "
                "no variation ".format(len(bad_cols))
            )

        if gwas_mode:
            _execute_gwas(
                genotypes, modelspec, subscribers, y.loc[idx, :], this_x,
                variant_predicates, messages, maf_t, cpus,
            )
        else:
            _execute_simple(
                modelspec, subscribers, y.loc[idx, :], this_x,
                variant_predicates, messages,
            )


def _get_uninformative_factors(df, modelspec):
    factor_cols = [
        i[2] for i in modelspec.transformations if i[0] == "ENCODE_FACTOR"
    ]

    bad_cols = []
    zero_cols = df.columns[df.sum() == 0]
    for col in zero_cols:
        for factor_col in factor_cols:
            if col.startswith(factor_col.id):
                bad_cols.append(col)
                break

    return bad_cols


def _execute_simple(modelspec, subscribers, y, X, variant_predicates,
                    messages):
    # There shouldn't be variant_predicates.
    if len(variant_predicates) != 0:
        logger.warning("Variant predicates are only used for GWAS "
                       "analyses.")

    # Get the statistical test.
    test = modelspec.test()

    logger.info(
        "Executing {} - Design matrix has shape: {}".format(test, X.shape)
    )

    # We don't need to worry about indexing or the sample order because
    # both parameters are from the same df.
    results = test.fit(y, X)

    # Update the results with the variant metadata.
    for entity in results:
        if entity in modelspec.variant_metadata:
            results[entity].update(modelspec.variant_metadata[entity])

    # Dispatch the results to the subscribers.
    for subscriber in subscribers:
        try:
            subscriber.handle(results)
        except KeyError as e:
            return subscribers_module.subscriber_error(e.args[0])


def _execute_gwas(genotypes, modelspec, subscribers, y, X, variant_predicates,
                  messages, maf_t=None, cpus=None):
        if cpus is None:
            cpus = multiprocessing.cpu_count() - 1

        # Create queues for failing SNPs and the consumer queue.
        failed = multiprocessing.Queue()
        q = multiprocessing.Queue(500)
        results = multiprocessing.Queue()
        abort = multiprocessing.Event()

        # Spawn the worker processes.
        workers = []
        for worker in range(cpus):
            this_y = y.copy()
            this_X = X.copy()
            fit = modelspec.test().fit
            samples = pd.Index(genotypes.get_samples())

            worker = multiprocessing.Process(
                target=_gwas_worker,
                args=(q, results, failed, abort, fit, this_y, this_X, samples,
                      maf_t)
            )

            workers.append(worker)
            worker.start()

        # Workers signal the end of their work by appending None to the results
        # queue. Hence, there should be as many Nones as workers in the
        # results queue by the end of the results processing.
        done_workers = 0

        def _handle_result():
            """Asynchronously processes an entry from the results queue.

            Does nothing if the queue is empty, passes through the subscribers
            if it is not.

            Returns 1 if there was a None in the queue, 0 otherwise. This is
            useful to track the number of workers that finish.

            """
            try:
                res = results.get(False)
            except queue.Empty:
                return 0

            if res is None:
                return 1

            for subscriber in subscribers:
                try:
                    subscriber.handle(res)
                except KeyError as e:
                    subscribers_module.subscriber_error(e.args[0], abort)

            return 0

        # Start filling the consumer queue and listening for results.
        not_analyzed = []
        for snp in genotypes.iter_genotypes():
            # Pass through the list of variant filtering predicates.
            try:
                if not all([f(snp) for f in variant_predicates]):
                    not_analyzed.append(snp.variant.name)
                    continue

            except StopIteration:
                break

            q.put(snp)

            # Handle results at the same time to avoid occupying too much
            # memory as the results queue gets filled.
            if abort.is_set():
                raise RuntimeError("Exception raised in worker processes")
            val = _handle_result()

            assert val == 0

        # Signal that there are no more SNPs to add.
        logger.debug("Done pushing SNPs")
        q.put(None)

        # Handle the remaining results.
        while done_workers != len(workers):
            done_workers += _handle_result()

        # Emptying the failed queue
        nb_failed_done = 0
        while nb_failed_done < cpus:
            snp = failed.get()
            if snp is None:
                nb_failed_done += 1
            else:
                messages["failed"].append(snp)
        failed.put(None)

        # Dump the not analyzed SNPs to disk.
        for snp in not_analyzed:
            messages["skipped"].append(snp)

        # Sanity check that there is nothing important left in the queues.
        queues_iter = zip(
            ('results', 'failed', 'q'), (results, failed, q)
        )
        for name, a_queue in queues_iter:
            while not a_queue.empty():
                val = a_queue.get()
                assert val is None, (name, val)

        # Join the workers.
        for worker in workers:
            worker.join()

        logger.info("Analysis completed")



"""
"""


# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


__all__ = ["PhenotypesContainer"]


class PhenotypesContainer(object):
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        raise NotImplementedError()

    def get_phenotypes(self, li=None):
        """Returns a dataframe of phenotypes.

        Args:
            li (list): A list of phenotypes to extract (default: None means
                       all phenotypes).

        Returns:
            pandas.DataFrame: A dataframe containing the phenotypes (with the
            sample IDs as index).

        """
        raise NotImplementedError()

    def get_nb_samples(self):
        """Returns the number of samples.

        Returns:
            int: The number of samples.

        """
        raise NotImplementedError()

    def get_nb_variables(self):
        """Returns the number of variables.

        Returns:
            int: The number of variables.

        """
        raise NotImplementedError()

    def is_repeated(self):
        """Check if the phenotypes contain repeated measurements.

        Returns:
            bool: ``True`` if the data contains repeated measurements,
            ``False`` otherwise.

        """
        raise NotImplementedError()

    def keep_samples(self, keep):
        """Keeps only a subset of samples.

        Args:
            keep (list): The list of samples to keep.

        """
        raise NotImplementedError()



"""
"""

# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import yaml

from geneparse import parsers as geno_map

from .statistics import available_models as available_tests
from .phenotypes import (
    available_formats as pheno_formats,
    format_map as pheno_map
)


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "CC BY-NC 4.0"


__all__ = ["AnalysisConfiguration"]


class AnalysisConfiguration(object):

    def __init__(self, filename):
        """Creates an instance of the class.

        Args:
            filename (str): The name of the configuration file.

        """
        # Reading the configuration file
        with open(filename, "r") as f:
            self._configuration = yaml.load(f)

        # Getting the required and available sections
        required_sections = {"genotypes", "phenotypes", "model"}
        available_sections = self._configuration.keys()

        # Checking if there are missing required sections
        missing_sections = required_sections - available_sections
        if len(missing_sections) > 0:
            raise ValueError(
                "Missing section(s) from the configuration file: {}.".format(
                    ", ".join(sorted(missing_sections)),
                )
            )

        # Checking if there are extra sections
        extra_sections = available_sections - required_sections
        if len(extra_sections) > 0:
            raise ValueError(
                "Invalid section(s) in the configuration file: {}.".format(
                    ", ".join(sorted(extra_sections)),
                )
            )

        # Configuring the Genotypes component
        self.configure_genotypes()

        # Configuring the Phenotypes component
        self.configure_phenotypes()

        # Configuring the model
        self.configure_model()

    def configure_genotypes(self):
        """Configures the genotypes component."""
        # Getting the genotype section of the configuration file
        section = self._configuration["genotypes"].copy()

        # Getting the format of the genotypes
        if "format" not in section:
            raise ValueError(
                "In the configuration file, the 'genotypes' section should "
                "contain the 'format' option specifying the format of the "
                "genotypes container."
            )
        geno_format = section.pop("format")

        # Checking if the format is valid
        if geno_format not in geno_map:
            raise ValueError(
                "Invalid 'format' ({}) for the 'genotypes' section of the "
                "configuration file.".format(geno_format)
            )

        # Getting the object to gather required and optional values
        self._geno_container = geno_map[geno_format]

        # We want to get the arguments for the genotype container
        self._geno_format = geno_format
        self._geno_arguments = self.retrieve_arguments(section)

    def get_genotypes_args(self):
        """Returns the genotypes arguments."""
        return self._geno_arguments

    def get_genotypes_format(self):
        """Returns the genotypes format."""
        return self._geno_format

    def get_genotypes(self):
        """Returns the genotypes container."""
        return self._geno_container(**self._geno_arguments)

    def configure_phenotypes(self):
        """Configures the phenotypes component."""
        # Getting the phenotype section of the configuration file
        section = self._configuration["phenotypes"].copy()

        # Getting the format of the phenotypes
        if "format" not in section:
            raise ValueError(
                "In the configuration file, the 'phenotypes' section should "
                "contain the 'format' option specifying the format of the "
                "phenotypes container."
            )
        pheno_format = section.pop("format")

        # Checking if the format is valid
        if pheno_format not in pheno_formats:
            raise ValueError(
                "Invalid 'format' ({}) for the 'phenotypes' section of the "
                "configuration file.".format(pheno_format)
            )

        # Getting the object to gather required and optional values
        self._pheno_container = pheno_map[pheno_format]

        # We want to get the arguments for the genotype container
        self._pheno_format = pheno_format
        self._pheno_arguments = self.retrieve_arguments(section)

    def get_phenotypes_args(self):
        """Returns the phenotypes arguments."""
        return self._pheno_arguments

    def get_phenotypes_format(self):
        """Returns the phenotypes format."""
        return self._pheno_format

    def get_phenotypes(self):
        """Returns the phenotypes container."""
        return self._pheno_container(**self._pheno_arguments)

    def configure_model(self):
        """Configures the statistics component."""
        # Getting the model section of the configuration file
        section = self._configuration["model"].copy()

        # Getting the required test
        if "test" not in section:
            raise ValueError(
                "In the configuration file, the 'model' section should "
                "contain the 'test' option specifying the statistical test "
                "to perform."
            )
        self._model_test = section.pop("test")

        # Checking the test is valid
        if self._model_test not in available_tests:
            raise ValueError(
                "Invalid 'test' for the 'model' section of the configuration "
                "file."
            )

        # Getting the formula
        if "formula" not in section:
            raise ValueError(
                "In the configuration file, the 'model' section should "
                "contain the 'formula' option specifying the statistical "
                "model to perform."
            )
        self._model_formula = section.pop("formula")

        # Getting the options
        self._model_arguments = self.retrieve_options(section)

    def get_model_args(self):
        """Returns the statistics arguments."""
        return self._model_arguments

    def get_model_test(self):
        """Returns the statistics model."""
        return self._model_test

    def get_model_formula(self):
        """Returns the model formula."""
        return self._model_formula

    @staticmethod
    def retrieve_arguments(config):
        """Retrieves arguments from the configuration.

        Args:
            config (configparser.SectionProxy): The section's configuration.

        Returns:
            dict: The arguments extracted from this section's configuration.

        """
        # The final arguments
        arguments = AnalysisConfiguration.retrieve_options(config)

        # Then, we gather the rest of the keys
        for key in config:
            arguments[key] = config[key]

        return arguments

    @staticmethod
    def retrieve_options(config):
        """Retrieves options from the configuration.

        Args:
            config (configparser.SectionProxy): The section's configuration.

        Returns:
            dict: The optional arguments.

        Note
        ====
            Only the restricted 'options' key is parsed, and has the following
            format: ``keyword=type:value`` (*e.g.* ``sep=str:,``). If there are
            multiple keyword/values, they are separated by a single coma.

        """
        options = {}

        if "options" in config:
            for k, v in config.pop("options").items():
                if v == r"\t":
                    v = "\t"
                options[k] = v

        return options



"""
Dummy implementation of the phenotype container with test data.
"""

# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import pandas as pd
import numpy as np

from .core import PhenotypesContainer


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


__all__ = ["_DummyPhenotypes"]


class _DummyPhenotypes(PhenotypesContainer):
    def __init__(self):
        self.data = pd.DataFrame({
            "y": [0, 1, 0, np.nan, 0],
            "V1": [1, 2, 3, 4, 5],
            "V2": [0, np.nan, 1, 0, 1],
            "V3": [-2, 1, np.nan, 0, -1],
            "V4": [0.1, 0.2, 0.3, 0.4, 0.5],
            "V5": [0, 0, 0, 1, 1],
        }, index=["s{}".format(i) for i in range(1, 6)])

    def close(self):
        pass

    def get_phenotypes(self, li=None):
        if li is None:
            li = self.data.columns

        for i in li:
            if i not in self.data.columns:
                raise KeyError(i)

        return self.data.loc[:, list(li)]

    def get_nb_samples(self):
        return self.data.shape[0]

    def get_nb_variables(self):
        return self.data.shape[1]

    def is_repeated(self):
        return False

    def keep_samples(self, keep):
        self.data = self.data.loc[keep, :]




# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.





# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import os
import unittest
from tempfile import TemporaryDirectory

import numpy as np
import pandas as pd
from pkg_resources import resource_filename

from pyplink import PyPlink
from geneparse.plink import PlinkReader

from ..statistics.core import StatsError
from ..statistics.models.linear import StatsLinear

from .. import analysis
from .. import subscribers
from .. import modelspec as spec
from ..phenotypes.dummy import _DummyPhenotypes


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


class TestStatsLinear(unittest.TestCase):
    """Tests the 'StatsLinear' class."""
    @classmethod
    def setUpClass(cls):
        # Loading the data
        cls.data = pd.read_csv(
            resource_filename(__name__, "data/statistics/linear.txt.bz2"),
            sep="\t",
            compression="bz2",
        )

        # Creating the index
        cls.data["sample"] = [
            "s{}".format(i+1) for i in range(cls.data.shape[0])
        ]
        cls.data = cls.data.set_index("sample")

        # Creating the dummy phenotype container
        cls.phenotypes = _DummyPhenotypes()
        cls.phenotypes.data = cls.data.drop(
            ["snp{}".format(i+1) for i in range(5)],
            axis=1,
        )

        # Creating a temporary directory
        cls.tmp_dir = TemporaryDirectory(prefix="genetest_test_linear_")

        # The plink file prefix
        cls.plink_prefix = os.path.join(cls.tmp_dir.name, "input")

        # Permuting the sample to add a bit of randomness
        new_sample_order = np.random.permutation(cls.data.index)

        # Creating the BED file
        with PyPlink(cls.plink_prefix, "w") as bed:
            for snp in [s for s in cls.data.columns if s.startswith("snp")]:
                bed.write_genotypes(cls.data.loc[new_sample_order, snp])

        # Creating the BIM file
        with open(cls.plink_prefix + ".bim", "w") as bim:
            print(3, "snp1", 0, 1234, "T", "C", sep="\t", file=bim)
            print(3, "snp2", 0, 9618, "C", "A", sep="\t", file=bim)
            print(2, "snp3", 0, 1519, "G", "T", sep="\t", file=bim)
            print(1, "snp4", 0, 5871, "G", "A", sep="\t", file=bim)
            print(23, "snp5", 0, 2938, "T", "C", sep="\t", file=bim)

        # Creating the FAM file
        with open(cls.plink_prefix + ".fam", "w") as fam:
            for sample in new_sample_order:
                print(sample, sample, 0, 0, 0, -9, file=fam)

        # Creating the genotype parser
        cls.genotypes = PlinkReader(cls.plink_prefix)

    @classmethod
    def tearDownClass(cls):
        cls.tmp_dir.cleanup()
        cls.genotypes.close()

    def setUp(self):
        # Resetting the model specification
        spec._reset()

        # Reordering the columns and the rows of the phenotype data frame
        self.phenotypes.data = self.phenotypes.data.iloc[
            np.random.permutation(self.phenotypes.data.shape[0]),
            np.random.permutation(self.phenotypes.data.shape[1])
        ]

    def test_linear_gwas(self):
        """Tests linear regression for GWAS."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno1,
            predictors=[spec.SNPs, spec.phenotypes.age,
                        spec.phenotypes.var1, gender],
            test="linear",
        )

        # The output prefix
        out_prefix = os.path.join(self.tmp_dir.name, "results")

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber], output_prefix=out_prefix,
        )
        gwas_results = subscriber._get_gwas_results()

        # Checking the number of results (should be 4)
        self.assertEqual(4, len(gwas_results.keys()))

        # The number of observations and parameters
        n = self.phenotypes.data.shape[0]
        p = 4

        # Checking the first marker (snp1)
        results = gwas_results["snp1"]
        self.assertEqual("snp1", results["SNPs"]["name"])
        self.assertEqual("3", results["SNPs"]["chrom"])
        self.assertEqual(1234, results["SNPs"]["pos"])
        self.assertEqual("T", results["SNPs"]["minor"])
        self.assertEqual("C", results["SNPs"]["major"])
        self.assertAlmostEqual(0.016666666666666666, results["SNPs"]["maf"])

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(113.19892138658, results["SNPs"]["coef"])
        self.assertAlmostEqual(20.8583649966504, results["SNPs"]["std_err"])
        self.assertAlmostEqual(71.397823827102, results["SNPs"]["lower_ci"])
        self.assertAlmostEqual(155.000018946058, results["SNPs"]["upper_ci"])
        self.assertAlmostEqual(5.42702754529217, results["SNPs"]["t_value"])
        self.assertAlmostEqual(-np.log10(0.0000013285915771),
                               -np.log10(results["SNPs"]["p_value"]))

        # Checking the model r squared (adjusted) (according to SAS)
        self.assertAlmostEqual(
            1 - (n - 1) * (1 - 0.35513322438349)/((n - 1) - p),
            results["MODEL"]["r_squared_adj"],
        )

        # Checking the second marker (snp2)
        results = gwas_results["snp2"]
        self.assertEqual("snp2", results["SNPs"]["name"])
        self.assertEqual("3", results["SNPs"]["chrom"])
        self.assertEqual(9618, results["SNPs"]["pos"])
        self.assertEqual("C", results["SNPs"]["minor"])
        self.assertEqual("A", results["SNPs"]["major"])
        self.assertAlmostEqual(0.20833333333333334, results["SNPs"]["maf"])

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(25.6638410624231, results["SNPs"]["coef"])
        self.assertAlmostEqual(7.02442421875627, results["SNPs"]["std_err"])
        self.assertAlmostEqual(11.5865803512147, results["SNPs"]["lower_ci"])
        self.assertAlmostEqual(39.7411017736316, results["SNPs"]["upper_ci"])
        self.assertAlmostEqual(3.65351525807579, results["SNPs"]["t_value"])
        self.assertAlmostEqual(-np.log10(0.0005783767026428),
                               -np.log10(results["SNPs"]["p_value"]))

        # Checking the model r squared (adjusted) (according to SAS)
        self.assertAlmostEqual(
            1 - (n - 1) * (1 - 0.20318728482079)/((n - 1) - p),
            results["MODEL"]["r_squared_adj"],
        )

        # Checking the third marker (snp3)
        results = gwas_results["snp3"]
        self.assertEqual("snp3", results["SNPs"]["name"])
        self.assertEqual("2", results["SNPs"]["chrom"])
        self.assertEqual(1519, results["SNPs"]["pos"])
        self.assertEqual("G", results["SNPs"]["minor"])
        self.assertEqual("T", results["SNPs"]["major"])
        self.assertAlmostEqual(0.29166666666666669, results["SNPs"]["maf"])

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(0.08097682855889, results["SNPs"]["coef"])
        self.assertAlmostEqual(6.6803747245602, results["SNPs"]["std_err"])
        self.assertAlmostEqual(-13.3067932886126, results["SNPs"]["lower_ci"])
        self.assertAlmostEqual(13.4687469457304, results["SNPs"]["upper_ci"])
        self.assertAlmostEqual(0.0121215997451737, results["SNPs"]["t_value"])
        self.assertAlmostEqual(0.99037246258077, results["SNPs"]["p_value"])

        # Checking the model r squared (adjusted) (according to SAS)
        self.assertAlmostEqual(
            1 - (n - 1) * (1 - 0.0098082108350667)/((n - 1) - p),
            results["MODEL"]["r_squared_adj"],
        )

        # Checking the fourth marker (snp4)
        results = gwas_results["snp4"]
        self.assertEqual("snp4", results["SNPs"]["name"])
        self.assertEqual("1", results["SNPs"]["chrom"])
        self.assertEqual(5871, results["SNPs"]["pos"])
        self.assertEqual("G", results["SNPs"]["minor"])
        self.assertEqual("A", results["SNPs"]["major"])
        self.assertAlmostEqual(0.275, results["SNPs"]["maf"])

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(-17.0933815760203, results["SNPs"]["coef"])
        self.assertAlmostEqual(6.49570434323821, results["SNPs"]["std_err"])
        self.assertAlmostEqual(-30.1110639788755, results["SNPs"]["lower_ci"])
        self.assertAlmostEqual(-4.07569917316514, results["SNPs"]["upper_ci"])
        self.assertAlmostEqual(-2.63149008526133, results["SNPs"]["t_value"])
        self.assertAlmostEqual(0.0110092290989312, results["SNPs"]["p_value"])

        # Checking the model r squared (adjusted) (according to SAS)
        self.assertAlmostEqual(
            1 - (n - 1) * (1 - 0.1205341542723)/((n - 1) - p),
            results["MODEL"]["r_squared_adj"],
        )

        # There should be a file for the failed snp5
        self.assertTrue(os.path.isfile(out_prefix + "_failed_snps.txt"))
        with open(out_prefix + "_failed_snps.txt") as f:
            self.assertEqual(
                [["snp5", "condition number is large, inf"]],
                [line.split("\t") for line in f.read().splitlines()],
            )

    def test_linear_snp1(self):
        """Tests linear regression with the first SNP."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno1,
            predictors=[spec.genotypes.snp1, spec.phenotypes.age,
                        spec.phenotypes.var1, gender],
            test="linear",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp1", results["snp1"]["name"])
        self.assertEqual("3", results["snp1"]["chrom"])
        self.assertEqual(1234, results["snp1"]["pos"])
        self.assertEqual("T", results["snp1"]["minor"])
        self.assertEqual("C", results["snp1"]["major"])
        self.assertAlmostEqual(0.016666666666666666, results["snp1"]["maf"])

        # The number of observations and parameters
        n = self.phenotypes.data.shape[0]
        p = 4

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(113.19892138658, results["snp1"]["coef"])
        self.assertAlmostEqual(20.8583649966504, results["snp1"]["std_err"])
        self.assertAlmostEqual(71.397823827102, results["snp1"]["lower_ci"])
        self.assertAlmostEqual(155.000018946058, results["snp1"]["upper_ci"])
        self.assertAlmostEqual(5.42702754529217, results["snp1"]["t_value"])
        self.assertAlmostEqual(-np.log10(0.0000013285915771),
                               -np.log10(results["snp1"]["p_value"]))

        # Checking the model r squared (adjusted) (according to SAS)
        self.assertAlmostEqual(
            1 - (n - 1) * (1 - 0.35513322438349)/((n - 1) - p),
            results["MODEL"]["r_squared_adj"],
        )

        # TODO: Check the other predictors

    def test_linear_snp1_inter(self):
        """Tests linear regression with the first SNP (interaction)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # The interaction term
        inter = spec.interaction(spec.phenotypes.var1, spec.genotypes.snp1)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno1,
            predictors=[spec.genotypes.snp1, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, inter],
            test=lambda: StatsLinear(condition_value_t=15000),
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp1", results["snp1"]["name"])
        self.assertEqual("3", results["snp1"]["chrom"])
        self.assertEqual(1234, results["snp1"]["pos"])
        self.assertEqual("T", results["snp1"]["minor"])
        self.assertEqual("C", results["snp1"]["major"])
        self.assertAlmostEqual(0.016666666666666666, results["snp1"]["maf"])

        # The number of observations and parameters
        n = self.phenotypes.data.shape[0]
        p = 5

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(28.3750067790686, results[inter.id]["coef"])
        self.assertAlmostEqual(15.31571903952, results[inter.id]["std_err"])
        self.assertAlmostEqual(-2.33116110697257,
                               results[inter.id]["lower_ci"])
        self.assertAlmostEqual(59.0811746651098,
                               results[inter.id]["upper_ci"])
        self.assertAlmostEqual(1.85267219291832,
                               results[inter.id]["t_value"])
        self.assertAlmostEqual(-np.log10(0.06939763567524),
                               -np.log10(results[inter.id]["p_value"]))

        # Checking the model r squared (adjusted) (according to SAS)
        self.assertAlmostEqual(
            1 - (n - 1) * (1 - 0.39367309450771)/((n - 1) - p),
            results["MODEL"]["r_squared_adj"],
        )

        # TODO: Check the other predictors

    def test_linear_snp1_inter_categorical(self):
        """Tests linear regression for first SNP (inter, category)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # The interaction term
        inter = spec.interaction(gender, spec.genotypes.snp1)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno1,
            predictors=[spec.genotypes.snp1, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, inter],
            test="linear",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp1", results["snp1"]["name"])
        self.assertEqual("3", results["snp1"]["chrom"])
        self.assertEqual(1234, results["snp1"]["pos"])
        self.assertEqual("T", results["snp1"]["minor"])
        self.assertEqual("C", results["snp1"]["major"])
        self.assertAlmostEqual(0.016666666666666666, results["snp1"]["maf"])

        # The number of observations and parameters
        n = self.phenotypes.data.shape[0]
        p = 5

        # Checking the marker statistics (according to SAS)
        col = inter.id + ":level.2"
        self.assertAlmostEqual(-74.5163756952978, results[col]["coef"])
        self.assertAlmostEqual(40.2210255975831, results[col]["std_err"])
        self.assertAlmostEqual(-155.154676865573, results[col]["lower_ci"],
                               places=6)
        self.assertAlmostEqual(6.12192547497808, results[col]["upper_ci"],
                               places=6)
        self.assertAlmostEqual(-1.8526721929183, results[col]["t_value"])
        self.assertAlmostEqual(-np.log10(0.06939763567525),
                               -np.log10(results[col]["p_value"]))

        # Checking the model r squared (adjusted) (according to SAS)
        self.assertAlmostEqual(
            1 - (n - 1) * (1 - 0.39367309450771)/((n - 1) - p),
            results["MODEL"]["r_squared_adj"],
        )

        # TODO: Check the other predictors

    def test_linear_snp2(self):
        """Tests linear regression with the second SNP."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno1,
            predictors=[spec.genotypes.snp2, spec.phenotypes.age,
                        spec.phenotypes.var1, gender],
            test="linear",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp2", results["snp2"]["name"])
        self.assertEqual("3", results["snp2"]["chrom"])
        self.assertEqual(9618, results["snp2"]["pos"])
        self.assertEqual("C", results["snp2"]["minor"])
        self.assertEqual("A", results["snp2"]["major"])
        self.assertAlmostEqual(0.20833333333333334, results["snp2"]["maf"])

        # The number of observations and parameters
        n = self.phenotypes.data.shape[0]
        p = 4

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(25.6638410624231, results["snp2"]["coef"])
        self.assertAlmostEqual(7.02442421875627, results["snp2"]["std_err"])
        self.assertAlmostEqual(11.5865803512147, results["snp2"]["lower_ci"])
        self.assertAlmostEqual(39.7411017736316, results["snp2"]["upper_ci"])
        self.assertAlmostEqual(3.65351525807579, results["snp2"]["t_value"])
        self.assertAlmostEqual(-np.log10(0.0005783767026428),
                               -np.log10(results["snp2"]["p_value"]))

        # Checking the model r squared (adjusted) (according to SAS)
        self.assertAlmostEqual(
            1 - (n - 1) * (1 - 0.20318728482079)/((n - 1) - p),
            results["MODEL"]["r_squared_adj"],
        )

        # TODO: Check the other predictors

    def test_linear_snp2_inter(self):
        """Tests linear regression with the second SNP (interaction)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # The interaction term
        inter = spec.interaction(spec.phenotypes.var1, spec.genotypes.snp2)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno1,
            predictors=[spec.genotypes.snp2, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, inter],
            test="linear",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp2", results["snp2"]["name"])
        self.assertEqual("3", results["snp2"]["chrom"])
        self.assertEqual(9618, results["snp2"]["pos"])
        self.assertEqual("C", results["snp2"]["minor"])
        self.assertEqual("A", results["snp2"]["major"])
        self.assertAlmostEqual(0.20833333333333334, results["snp2"]["maf"])

        # The number of observations and parameters
        n = self.phenotypes.data.shape[0]
        p = 5

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(-0.38040787976905, results[inter.id]["coef"])
        self.assertAlmostEqual(0.56827855931761, results[inter.id]["std_err"])
        self.assertAlmostEqual(-1.5197377932663, results[inter.id]["lower_ci"])
        self.assertAlmostEqual(0.75892203372818, results[inter.id]["upper_ci"])
        self.assertAlmostEqual(-0.66940389274205, results[inter.id]["t_value"])
        self.assertAlmostEqual(-np.log10(0.50609004475028),
                               -np.log10(results[inter.id]["p_value"]))

        # Checking the model r squared (adjusted) (according to SAS)
        self.assertAlmostEqual(
            1 - (n - 1) * (1 - 0.20974496120713)/((n - 1) - p),
            results["MODEL"]["r_squared_adj"],
        )

        # TODO: Check the other predictors

    def test_linear_snp3(self):
        """Tests linear regression with the third SNP."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno1,
            predictors=[spec.genotypes.snp3, spec.phenotypes.age,
                        spec.phenotypes.var1, gender],
            test="linear",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp3", results["snp3"]["name"])
        self.assertEqual("2", results["snp3"]["chrom"])
        self.assertEqual(1519, results["snp3"]["pos"])
        self.assertEqual("G", results["snp3"]["minor"])
        self.assertEqual("T", results["snp3"]["major"])
        self.assertAlmostEqual(0.29166666666666669, results["snp3"]["maf"])

        # The number of observations and parameters
        n = self.phenotypes.data.shape[0]
        p = 4

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(0.08097682855889, results["snp3"]["coef"])
        self.assertAlmostEqual(6.6803747245602, results["snp3"]["std_err"])
        self.assertAlmostEqual(-13.3067932886126, results["snp3"]["lower_ci"])
        self.assertAlmostEqual(13.4687469457304, results["snp3"]["upper_ci"])
        self.assertAlmostEqual(0.0121215997451737, results["snp3"]["t_value"])
        self.assertAlmostEqual(0.99037246258077, results["snp3"]["p_value"])

        # Checking the model r squared (adjusted) (according to SAS)
        self.assertAlmostEqual(
            1 - (n - 1) * (1 - 0.0098082108350667)/((n - 1) - p),
            results["MODEL"]["r_squared_adj"],
        )

        # TODO: Check the other predictors

    def test_linear_snp3_inter(self):
        """Tests linear regression with the third SNP."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # The interaction term
        inter = spec.interaction(spec.phenotypes.var1, spec.genotypes.snp3)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno1,
            predictors=[spec.genotypes.snp3, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, inter],
            test="linear",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp3", results["snp3"]["name"])
        self.assertEqual("2", results["snp3"]["chrom"])
        self.assertEqual(1519, results["snp3"]["pos"])
        self.assertEqual("G", results["snp3"]["minor"])
        self.assertEqual("T", results["snp3"]["major"])
        self.assertAlmostEqual(0.29166666666666669, results["snp3"]["maf"])

        # The number of observations and parameters
        n = self.phenotypes.data.shape[0]
        p = 5

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(-0.0097102715733324, results[inter.id]["coef"])
        self.assertAlmostEqual(0.60510302626961, results[inter.id]["std_err"])
        self.assertAlmostEqual(-1.22286879616119,
                               results[inter.id]["lower_ci"])
        self.assertAlmostEqual(1.20344825301452, results[inter.id]["upper_ci"])
        self.assertAlmostEqual(-0.0160473029414429,
                               results[inter.id]["t_value"])
        self.assertAlmostEqual(0.98725579876123, results[inter.id]["p_value"])

        # Checking the model r squared (adjusted) (according to SAS)
        self.assertAlmostEqual(
            1 - (n - 1) * (1 - 0.0098129328525696)/((n - 1) - p),
            results["MODEL"]["r_squared_adj"],
        )

        # TODO: Check the other predictors

    def test_linear_snp4(self):
        """Tests linear regression with the fourth SNP."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno1,
            predictors=[spec.genotypes.snp4, spec.phenotypes.age,
                        spec.phenotypes.var1, gender],
            test="linear",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp4", results["snp4"]["name"])
        self.assertEqual("1", results["snp4"]["chrom"])
        self.assertEqual(5871, results["snp4"]["pos"])
        self.assertEqual("G", results["snp4"]["minor"])
        self.assertEqual("A", results["snp4"]["major"])
        self.assertAlmostEqual(0.275, results["snp4"]["maf"])

        # The number of observations and parameters
        n = self.phenotypes.data.shape[0]
        p = 4

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(-17.0933815760203, results["snp4"]["coef"])
        self.assertAlmostEqual(6.49570434323821, results["snp4"]["std_err"])
        self.assertAlmostEqual(-30.1110639788755, results["snp4"]["lower_ci"])
        self.assertAlmostEqual(-4.07569917316514, results["snp4"]["upper_ci"])
        self.assertAlmostEqual(-2.63149008526133, results["snp4"]["t_value"])
        self.assertAlmostEqual(0.0110092290989312, results["snp4"]["p_value"])

        # Checking the model r squared (adjusted) (according to SAS)
        self.assertAlmostEqual(
            1 - (n - 1) * (1 - 0.1205341542723)/((n - 1) - p),
            results["MODEL"]["r_squared_adj"],
        )

        # TODO: Check the other predictors

    def test_linear_snp4_inter(self):
        """Tests linear regression with the fourth SNP (interaction)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # The interaction term
        inter = spec.interaction(spec.phenotypes.var1, spec.genotypes.snp4)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno1,
            predictors=[spec.genotypes.snp4, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, inter],
            test="linear",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp4", results["snp4"]["name"])
        self.assertEqual("1", results["snp4"]["chrom"])
        self.assertEqual(5871, results["snp4"]["pos"])
        self.assertEqual("G", results["snp4"]["minor"])
        self.assertEqual("A", results["snp4"]["major"])
        self.assertAlmostEqual(0.275, results["snp4"]["maf"])

        # The number of observations and parameters
        n = self.phenotypes.data.shape[0]
        p = 5

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(-0.46834820683113, results[inter.id]["coef"])
        self.assertAlmostEqual(0.50898831048606, results[inter.id]["std_err"])
        self.assertAlmostEqual(-1.48880832845448,
                               results[inter.id]["lower_ci"])
        self.assertAlmostEqual(0.5521119147922, results[inter.id]["upper_ci"])
        self.assertAlmostEqual(-0.92015513359016, results[inter.id]["t_value"])
        self.assertAlmostEqual(0.36158411106165, results[inter.id]["p_value"])

        # Checking the model r squared (adjusted) (according to SAS)
        self.assertAlmostEqual(
            1 - (n - 1) * (1 - 0.13411074411446)/((n - 1) - p),
            results["MODEL"]["r_squared_adj"],
        )

        # TODO: Check the other predictors

    def test_linear_snp5(self):
        """Tests linear regression with the fifth SNP (raises StatsError)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno1,
            predictors=[spec.genotypes.snp5, spec.phenotypes.age,
                        spec.phenotypes.var1, gender],
            test="linear",
        )

        # Performing the analysis and retrieving the results
        with self.assertRaises(StatsError) as cm:
            analysis.execute(
                self.phenotypes, self.genotypes, modelspec,
                subscribers=[subscribers.ResultsMemory()],
            )
        self.assertEqual("condition number is large, inf", str(cm.exception))

    def test_linear_snp5_inter(self):
        """Tests linear regression fifth SNP (inter) (raises StatsError)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # The interaction term
        inter = spec.interaction(spec.phenotypes.var1, spec.genotypes.snp5)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno1,
            predictors=[spec.genotypes.snp5, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, inter],
            test="linear",
        )

        # Performing the analysis and retrieving the results
        with self.assertRaises(StatsError) as cm:
            analysis.execute(
                self.phenotypes, self.genotypes, modelspec,
                subscribers=[subscribers.ResultsMemory()],
            )
        self.assertEqual("condition number is large, inf", str(cm.exception))



"""
Utilities to build statistical models.
"""

# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import re
import uuid
import operator
import itertools
import functools

import numpy as np
import pandas as pd

from ..statistics import model_map
from ..statistics.descriptive import get_maf


SNPs = "SNPs"
model = "MODEL"


class PheWAS(object):
    """A collection of EntityIdentifiers used to describe pheWAS analyses."""
    def __init__(self, li=None):
        self.li = li


class transformation_handler(object):
    handlers = {}

    def __init__(self, action):
        self.action = action

    def __call__(self, f):
        self.__class__.handlers[self.action] = f
        return f


class Result(object):
    def __init__(self, entity):
        self.path = [entity]

    def __getitem__(self, key):
        self.path.append(key)
        return self

    def get(self, results):
        cur = results

        for field in self.path:
            if isinstance(field, EntityIdentifier):
                # Access for transformations and entities.
                if field.id in results.keys():
                    cur = cur[field.id]
                else:
                    # Check for many levels.
                    results_per_level = {}

                    for key in results:
                        split_key = key.split(":")
                        id, level = (":".join(split_key[:-1]), split_key[-1])
                        if id == field.id:
                            results_per_level[level] = cur[key]

                    cur = results_per_level

            else:
                # Access for special fields like "SNPs", "MODEL" or for
                # regular keys.
                cur = cur[field]

        return cur


class ResultMetaclass(object):
    def __getitem__(self, entity):
        return Result(entity)


class Expression(object):
    def __init__(self, expression):
        self.expression = expression

    def __mul__(self, other):
        return Expression((operator.mul, self.expression, other))

    def __div__(self, other):
        return Expression((operator.div, self.expression, other))

    def __add__(self, other):
        return Expression((operator.add, self.expression, other))

    def __sub__(self, other):
        return Expression((operator.sub, self.expression, other))

    __rmul__ = __mul__
    __rdiv__ = __div__
    __radd__ = __add__
    __rsub__ = __sub__

    def eval(self):
        return self._eval_expression(self.expression)

    @staticmethod
    def _eval_expression(expression):
        op, left, right = expression

        if isinstance(left, tuple):
            left = Expression._eval_expression(left)

        if isinstance(left, Expression):
            left = Expression._eval_expression(left.expression)

        if isinstance(right, tuple):
            right = Expression._eval_expression(right)

        if isinstance(right, Expression):
            right = Expression._eval_expression(right.expression)

        if hasattr(left, "values"):
            left = left.values

        if hasattr(right, "values"):
            right = right.values

        return op(left, right)


class EntityIdentifier(object):
    """Placeholder for variables in the modelspec.

    Instances get assigned an ID (a UUID4 if nothing is specified), the
    chosen identifier otherwise.

    It is possible to do basic arithmetic on instances. This will return an
    Expression object that can be evaluated. Before evaluation, data needs
    to be bound to the EntityIdentifier using the bind method.

    """
    def __init__(self, id=None):
        if id:
            self.id = id
        else:
            self.id = str(uuid.uuid4())

        self.values = None

    def __mul__(self, other):
        return Expression((operator.mul, self, other))

    def __div__(self, other):
        return Expression((operator.div, self, other))

    def __add__(self, other):
        return Expression((operator.add, self, other))

    def __sub__(self, other):
        return Expression((operator.sub, self, other))

    __rmul__ = __mul__
    __rdiv__ = __div__
    __radd__ = __add__
    __rsub__ = __sub__

    def __repr__(self):
        return self.id.split("-")[0]

    def bind(self, values):
        self.values = values


class DependencyManager(object):
    """Class that remembers which items are accessed in an internal set.

    For example, "genotypes" and "phenotypes" are dependency managers.

    """
    dependencies = {}

    def __init__(self, source):
        self.source = source

    def __getitem__(self, key):
        deps = self.__class__.dependencies

        if (self.source, key) in deps:
            return deps[(self.source, key)]

        id = EntityIdentifier()
        deps[(self.source, key)] = id
        return id

    __getattr__ = __getitem__


class TransformationManager(object):
    """Class that remembers which data manipulation will be necessary."""
    transformations = []

    def __init__(self, action):
        self.action = action

    def __call__(self, source, *params, name=None):
        # source is either an EntityIdentifier or an Expression
        # Generate an identifier for the result of the transformation.
        if name is None:
            # Generate an interpretable name.
            if isinstance(source, Expression):
                raise ValueError(
                    "The name parameter is mandatory for expression-based "
                    "transformations."
                )
            else:
                name = "TRANSFORM:{}:{}".format(self.action, source.id)

        else:
            name = "TRANSFORM:{}".format(name)

        target = EntityIdentifier(name)

        # Make sure there are no entities with the same name.
        existing_transformations = {
            i[2].id: i[2] for i in TransformationManager.transformations
        }
        if target.id in existing_transformations:
            # The transformation was already performed, so we'll just return
            # the already created target
            return existing_transformations[target.id]

        TransformationManager.transformations.append(
            (self.action, source, target, params)
        )
        return target


class ModelSpec(object):
    def __init__(self, outcome, predictors, test, no_intercept=False,
                 stratify_by=None):
        """Statistical model specification.

        Args:
            outcome (EntityIdentifier): The outcome variable of the model.
            predictors (list): A list of EntityIdentifier that represent
                               covariates.
            test (callable or str): Either the name of a statistical test or
                                    a callable that returns an instance of a
                                    statistical test.
            no_intercept (bool): Controls if a column of ones is to be added.
            stratify_by (list): A list of EntityIdentifier instances
                                to stratify the analysis.

        """
        self.outcome = self._clean_outcome(outcome)
        self.predictors = self._clean_predictors(predictors)
        self.test = self._clean_test(test)
        self.no_intercept = no_intercept

        if hasattr(stratify_by, "__iter__"):
            self.stratify_by = stratify_by
        elif stratify_by is not None:
            self.stratify_by = list(stratify_by)
        else:
            self.stratify_by = None

        # SNP metadata is stored in the modelspec because it is obtained as a
        # consequence of building the data matrix.
        # Hence, it is more efficient to cache it than to request it.
        self.variant_metadata = {}

    def _clean_test(self, test):
        """Returns a factory function to create instances of a statistical
        model.

        Args:
            test (str, class or callable): This can be either the name of the
                                           test or a class or callable that
                                           returns a valid instance.

        """
        try:
            return model_map[test]
        except KeyError:
            pass

        if hasattr(test, "__call__"):
            return test

        raise ValueError(
            "{} is not a valid statistical model.".format(test)
        )

    def _clean_outcome(self, outcome):
        if isinstance(outcome, EntityIdentifier):
            return {"y": outcome}

        return outcome

    def _clean_predictors(self, predictors):
        iter_failed = True
        try:
            iter(predictors)
            iter_failed = False
        except TypeError:
            pass

        if iter_failed:
            raise TypeError("'predictors' argument needs to be an iterable.")

        return predictors

    @property
    def dependencies(self):
        return DependencyManager.dependencies

    @property
    def transformations(self):
        return TransformationManager.transformations

    def get_tested_variants(self):
        if SNPs in self.predictors:
            return SNPs
        else:
            return [
                (v, k[1]) for k, v in self.dependencies.items()
                if k[0] == "GENOTYPES"
            ]

    def get_translations(self):
        """Returns a dict mapping IDs to regular variable names."""
        return {id.id: tu[1] for tu, id in self.dependencies.items()}

    def create_data_matrix(self, phenotypes, genotypes):
        """Create the data matrix given data containers.

        Args:
            phenotypes (PhenotypeContainer): The phenotypes.
            genotypes (geneparse.core.Genotypes): The genotypes.

        """
        PHENOTYPES = "PHENOTYPES"

        # Extract the phenotype dependencies.
        if isinstance(self.outcome, PheWAS) and self.outcome.li is None:
            # Get all the phenotypes.
            df = phenotypes.get_phenotypes()

            # Create entities if needed for the outcomes.
            self.outcome.li = []

            col_names = []
            for col in df.columns:
                # Get an entity if it already exists.
                entity = self.dependencies.get((PHENOTYPES, col))

                # Create the entity if it didn't exist.
                if entity is None:
                    entity = EntityIdentifier()
                    self.dependencies[(PHENOTYPES, col)] = entity

                # If the entity is not in the predictors, we should analyze it
                # in the pheWAS.
                if entity not in self.predictors:
                    self.outcome.li.append(entity)

                col_names.append(entity.id)

            df.columns = col_names

        else:
            phen_keys = [
                k[1] for k, v in self.dependencies.items()
                if k[0] == PHENOTYPES
            ]

            df = phenotypes.get_phenotypes(phen_keys)
            df.columns = [
                self.dependencies[(PHENOTYPES, k)].id for k in phen_keys
            ]

        # Extract the genotype dependencies.
        markers = self.get_tested_variants()

        # In GWAS we handle the SNPs differently, but the modelspec adds
        # variants that are specified explicitly directly to the phenotypes
        # dataframe.
        if markers is not SNPs:
            for entity, marker in markers:
                entity_id = self.dependencies[("GENOTYPES", marker)]

                try:
                    g = genotypes.get_variant_by_name(marker)
                except (KeyError, ValueError):
                    raise ValueError(
                        "Could not find '{}' in genotypes container."
                        "".format(marker)
                    )

                # Checking we have a single value and using it
                if len(g) != 1:
                    raise ValueError("{}: invalid marker name".format(marker))
                g = g.pop()

                # Rename the genotypes column before joining.
                df = df.join(
                    pd.Series(g.genotypes, index=genotypes.get_samples(),
                              name=entity_id.id),
                    how="inner",
                )

                if df.shape[0] == 0:
                    raise ValueError(
                        "No sample left after joining. Perhaps the sample IDs "
                        "in the genotypes and phenotypes containers are "
                        "different."
                    )

                # Compute the maf.
                maf, minor, major, flip = get_maf(
                    df[entity_id.id], g.coded, g.reference,
                )

                if flip:
                    df.loc[:, entity_id.id] = 2 - df.loc[:, entity_id.id]

                # Also bind the EntityIdentifier in case we need to compute
                # a GRS.
                entity_id.bind(df[entity_id.id])

                # And save the variant metadata.
                self.variant_metadata[entity.id] = {
                    "name": marker, "chrom": g.variant.chrom, "pos":
                    g.variant.pos, "minor": minor, "major": major, "maf": maf
                }

        # Apply transformations.
        df = self._apply_transformations(df)

        # Only keep predictors and outcomes.
        keep_cols = self._filter_columns(df)

        # Adding the intercept
        if not self.no_intercept:
            keep_cols.append("intercept")
            df["intercept"] = 1

        return df[keep_cols]

    def _apply_transformations(self, df):
        for action, source, target, params in self.transformations:
            f = transformation_handler.handlers[action]
            res = f(df, source, *params)

            # Some transformations return multiple columns. We create all the
            # relevant columns in the dataframe.
            if isinstance(res, dict):
                for key, col in res.items():
                    df["{}:{}".format(target.id, key)] = col

            # In most cases, transformations return a single array. We set it
            # under the target ID.
            else:
                df[target.id] = res

        return df

    def _filter_columns(self, df):
        if isinstance(self.outcome, PheWAS):
            keep_cols = [
                v.id for v in self.outcome.li if v not in self.predictors
            ]
        else:
            keep_cols = [v.id for v in self.outcome.values()]

        for pred in self.predictors:
            if pred is SNPs:
                continue

            if not isinstance(pred, EntityIdentifier):
                raise ValueError(
                    "Predictors are expected to be entity identifiers (and "
                    "'{}' is of type {}).".format(pred, type(pred))
                )

            # Some identifiers might have multiple associated columns (e.g.
            # factor transformation).
            for col in df.columns:
                if col.startswith(pred.id):
                    keep_cols.append(col)

        if self.stratify_by is not None:
            for col in self.stratify_by:
                if not isinstance(col, EntityIdentifier):
                    raise ValueError(
                        "Statification variables are expected to be entity "
                        "identifiers (and '{}' is of type {})."
                        "".format(col, type(col))
                    )
                keep_cols.append(col.id)

        return keep_cols


@transformation_handler("LOG10")
def _log10(data, entity):
    return np.log10(data[entity.id])


@transformation_handler("LN")
def _ln(data, entity):
    return np.log(data[entity.id])


@transformation_handler("ENCODE_FACTOR")
def _encode_factor(data, entity):
    out = {}

    v = data[entity.id]
    nulls = v.isnull()

    # Pandas category.
    levels = None
    if hasattr(v, "cat"):
        levels = sorted(v.cat.categories)
    else:
        levels = sorted(v[~nulls].unique())

    for i, level in enumerate(levels):
        # First level is the reference.
        if i == 0:
            continue
        level_name = "level.{}".format(level)
        out[level_name] = (v == level).astype(float)
        out[level_name][nulls] = np.nan

    return out


@transformation_handler("POW")
def _pow(data, entity, power):
    return np.power(data[entity.id], power)


@transformation_handler("INTERACTION")
def _interaction(data, *entities):
    # Finding all the columns for all the targets
    column_names = tuple(
        tuple(name for name in data.columns if name.startswith(entity.id))
        for entity in entities
    )

    # Finding the level column names if there are factors
    factor_levels = tuple(
        tuple(name[len(entity.id)+1:] for name in names)
        for names, entity in zip(column_names, entities)
    )

    # Creating the final columns
    out = {}
    for cols, level_names in zip(itertools.product(*column_names),
                                 itertools.product(*factor_levels)):
        # Getting the key (for factors, if present)
        key = re.sub(":{2,}", "", ":".join(level_names).strip(":"))

        # Computing the multiplication of all the columns
        mult = functools.reduce(np.multiply, (data[col] for col in cols))

        # If the key is empty, it means there are no factors in the interaction
        # term, hence there is only one resulting column and we return it
        if key == "":
            return mult

        out[key] = mult

    return out


def _reset():
    TransformationManager.transformations = []
    DependencyManager.dependencies = {}


result = ResultMetaclass()

phenotypes = DependencyManager("PHENOTYPES")
genotypes = DependencyManager("GENOTYPES")

factor = TransformationManager("ENCODE_FACTOR")
log10 = TransformationManager("LOG10")
ln = TransformationManager("LN")
pow = TransformationManager("POW")
interaction = TransformationManager("INTERACTION")





# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import unittest

import numpy as np
import pandas as pd
from pkg_resources import resource_filename

from geneparse.dataframe import DataFrameReader

from ..statistics.models.mixedlm import StatsMixedLM

from .. import analysis
from .. import subscribers
from .. import modelspec as spec
from ..phenotypes.dummy import _DummyPhenotypes


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


class TestStatsMixedLM(unittest.TestCase):
    """Tests the 'StatsMixedLM' class."""
    @classmethod
    def setUpClass(cls):
        # Loading the data
        cls.data = pd.read_csv(
            resource_filename(__name__, "data/statistics/mixedlm.txt.bz2"),
            sep="\t",
            compression="bz2",
        )

        # Creating the index
        cls.data = cls.data.set_index("sampleid", drop=False)

        # Creating the dummy phenotype container
        cls.phenotypes = _DummyPhenotypes()
        cls.phenotypes.data = cls.data.drop(
            [col for col in cls.data.columns if col.startswith("snp")],
            axis=1,
        )

        # Permuting the sample to add a bit of randomness
        new_sample_order = np.random.permutation(cls.data.index)

        # Creating the genotypes data frame
        genotypes = cls.data.loc[
            new_sample_order,
            [col for col in cls.data.columns if col.startswith("snp")],
        ].copy()

        # Keeping only one copy of each data
        genotypes = genotypes[~genotypes.index.duplicated()]

        # Creating the mapping information
        map_info = pd.DataFrame(
            {"chrom": ["3", "3", "2", "22"],
             "pos": [1234, 2345, 3456, 4567],
             "a1": ["T", "C", "G", "A"],
             "a2": ["G", "A", "T", "C"]},
            index=["snp1", "snp2", "snp3", "snp4"],
        )

        # Creating the genotype parser
        cls.genotypes = DataFrameReader(
            dataframe=genotypes,
            map_info=map_info,
        )

    def setUp(self):
        # Resetting the model specification
        spec._reset()

        # Reordering the columns and the rows of the phenotypes data frame
        self.phenotypes.data = self.phenotypes.data.iloc[
            np.random.permutation(self.phenotypes.data.shape[0]),
            np.random.permutation(self.phenotypes.data.shape[1])
        ]

    def test_mixedlm_gwas(self):
        """Tests mixedlm regression for GWAS."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)
        visit = spec.factor(spec.phenotypes.visit)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome={"outcome": spec.phenotypes.pheno3,
                     "groups": spec.phenotypes.sampleid},
            predictors=[spec.SNPs, spec.phenotypes.age, spec.phenotypes.var1,
                        gender, visit],
            test="mixedlm",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        gwas_results = subscriber._get_gwas_results()

        # Checking the number of results (should be 4)
        self.assertEqual(4, len(gwas_results.keys()))

        # Checking the first marker (snp1)
        results = gwas_results["snp1"]
        self.assertEqual("snp1", results["SNPs"]["name"])
        self.assertEqual("3", results["SNPs"]["chrom"])
        self.assertEqual(1234, results["SNPs"]["pos"])
        self.assertEqual("T", results["SNPs"]["minor"])
        self.assertEqual("G", results["SNPs"]["major"])
        self.assertAlmostEqual(0.17499166666666666, results["SNPs"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(37.45704065571509034, results["SNPs"]["coef"])
        self.assertAlmostEqual(8.3417880825867652, results["SNPs"]["std_err"],
                               places=4)
        self.assertAlmostEqual(21.1074364471796017,
                               results["SNPs"]["lower_ci"], places=4)
        self.assertAlmostEqual(53.806644864250579,
                               results["SNPs"]["upper_ci"], places=4)
        self.assertAlmostEqual(4.4902891664085249,
                               results["SNPs"]["z_value"], places=4)
        self.assertAlmostEqual(-np.log10(7.112654716978639e-06),
                               -np.log10(results["SNPs"]["p_value"]), places=4)

        # Checking the second marker (snp2)
        results = gwas_results["snp2"]
        self.assertEqual("snp2", results["SNPs"]["name"])
        self.assertEqual("3", results["SNPs"]["chrom"])
        self.assertEqual(2345, results["SNPs"]["pos"])
        self.assertEqual("C", results["SNPs"]["minor"])
        self.assertEqual("A", results["SNPs"]["major"])
        self.assertAlmostEqual(0.36666666666666664, results["SNPs"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(28.87619649886422479, results["SNPs"]["coef"])
        self.assertAlmostEqual(7.616420586507464, results["SNPs"]["std_err"],
                               places=4)
        self.assertAlmostEqual(13.9482864582001636,
                               results["SNPs"]["lower_ci"], places=4)
        self.assertAlmostEqual(43.8041065395282843,
                               results["SNPs"]["upper_ci"], places=4)
        self.assertAlmostEqual(3.7913080259798924,
                               results["SNPs"]["z_value"], places=4)
        self.assertAlmostEqual(-np.log10(1.498559584953707e-04),
                               -np.log10(results["SNPs"]["p_value"]), places=4)

        # Checking the third marker (snp3)
        results = gwas_results["snp3"]
        self.assertEqual("snp3", results["SNPs"]["name"])
        self.assertEqual("2", results["SNPs"]["chrom"])
        self.assertEqual(3456, results["SNPs"]["pos"])
        self.assertEqual("G", results["SNPs"]["minor"])
        self.assertEqual("T", results["SNPs"]["major"])
        self.assertAlmostEqual(0.40833333333333333, results["SNPs"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(21.61350866000438486, results["SNPs"]["coef"])
        self.assertAlmostEqual(6.4018199962254876, results["SNPs"]["std_err"],
                               places=4)
        self.assertAlmostEqual(9.0661720318940873,
                               results["SNPs"]["lower_ci"], places=4)
        self.assertAlmostEqual(34.160845288114686,
                               results["SNPs"]["upper_ci"], places=4)
        self.assertAlmostEqual(3.37615063727935283,
                               results["SNPs"]["z_value"], places=4)
        self.assertAlmostEqual(-np.log10(7.350766113720653e-04),
                               -np.log10(results["SNPs"]["p_value"]), places=4)

        # Checking the fourth marker (snp4)
        results = gwas_results["snp4"]
        self.assertEqual("snp4", results["SNPs"]["name"])
        self.assertEqual("22", results["SNPs"]["chrom"])
        self.assertEqual(4567, results["SNPs"]["pos"])
        self.assertEqual("A", results["SNPs"]["minor"])
        self.assertEqual("C", results["SNPs"]["major"])
        self.assertAlmostEqual(0.44166666666666665, results["SNPs"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(1.4891822461435404, results["SNPs"]["coef"])
        self.assertAlmostEqual(7.9840787818167422, results["SNPs"]["std_err"],
                               places=3)
        self.assertAlmostEqual(-14.1593246159476980,
                               results["SNPs"]["lower_ci"], places=3)
        self.assertAlmostEqual(17.1376891082347775,
                               results["SNPs"]["upper_ci"], places=3)
        self.assertAlmostEqual(0.1865189819437983,
                               results["SNPs"]["z_value"], places=5)
        self.assertAlmostEqual(-np.log10(8.520377946017625e-01),
                               -np.log10(results["SNPs"]["p_value"]), places=5)

    def test_mixedlm_snp1_reml(self):
        """Tests mixedlm regression with the first SNP (using REML)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)
        visit = spec.factor(spec.phenotypes.visit)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome={"outcome": spec.phenotypes.pheno3,
                     "groups": spec.phenotypes.sampleid},
            predictors=[spec.genotypes.snp1, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, visit],
            test="mixedlm",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp1", results["snp1"]["name"])
        self.assertEqual("3", results["snp1"]["chrom"])
        self.assertEqual(1234, results["snp1"]["pos"])
        self.assertEqual("T", results["snp1"]["minor"])
        self.assertEqual("G", results["snp1"]["major"])
        self.assertAlmostEqual(0.17499166666666666, results["snp1"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(37.45704065571509034, results["snp1"]["coef"])
        self.assertAlmostEqual(8.3417880825867652, results["snp1"]["std_err"],
                               places=4)
        self.assertAlmostEqual(21.1074364471796017,
                               results["snp1"]["lower_ci"], places=4)
        self.assertAlmostEqual(53.806644864250579,
                               results["snp1"]["upper_ci"], places=4)
        self.assertAlmostEqual(4.4902891664085249,
                               results["snp1"]["z_value"], places=4)
        self.assertAlmostEqual(-np.log10(7.112654716978639e-06),
                               -np.log10(results["snp1"]["p_value"]), places=4)

        # TODO: Check the other predictors

    def test_mixedlm_snp1_ml(self):
        """Tests mixedlm regression with the first SNP (using ML)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)
        visit = spec.factor(spec.phenotypes.visit)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome={"outcome": spec.phenotypes.pheno3,
                     "groups": spec.phenotypes.sampleid},
            predictors=[spec.genotypes.snp1, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, visit],
            test=lambda: StatsMixedLM(reml=False),
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp1", results["snp1"]["name"])
        self.assertEqual("3", results["snp1"]["chrom"])
        self.assertEqual(1234, results["snp1"]["pos"])
        self.assertEqual("T", results["snp1"]["minor"])
        self.assertEqual("G", results["snp1"]["major"])
        self.assertAlmostEqual(0.17499166666666666, results["snp1"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(37.45704065571743513, results["snp1"]["coef"])
        self.assertAlmostEqual(7.986654823996228, results["snp1"]["std_err"],
                               places=5)
        self.assertAlmostEqual(21.8034848437317450,
                               results["snp1"]["lower_ci"], places=4)
        self.assertAlmostEqual(53.1105964677031253,
                               results["snp1"]["upper_ci"], places=4)
        self.assertAlmostEqual(4.6899536140182541,
                               results["snp1"]["z_value"], places=5)
        self.assertAlmostEqual(-np.log10(2.732669904137452e-06),
                               -np.log10(results["snp1"]["p_value"]), places=5)

        # TODO: Check the other predictors

    def test_mixedlm_snp1_inter_reml(self):
        """Tests mixedlm regression with the first SNP (using REML, inter)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)
        visit = spec.factor(spec.phenotypes.visit)

        # The interaction term
        inter = spec.interaction(spec.genotypes.snp1, spec.phenotypes.var1)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome={"outcome": spec.phenotypes.pheno3,
                     "groups": spec.phenotypes.sampleid},
            predictors=[spec.genotypes.snp1, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, visit, inter],
            test="mixedlm",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp1", results["snp1"]["name"])
        self.assertEqual("3", results["snp1"]["chrom"])
        self.assertEqual(1234, results["snp1"]["pos"])
        self.assertEqual("T", results["snp1"]["minor"])
        self.assertEqual("G", results["snp1"]["major"])
        self.assertAlmostEqual(0.17499166666666666, results["snp1"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(1.357502192968099, results[inter.id]["coef"])
        self.assertAlmostEqual(0.6710947924603057,
                               results[inter.id]["std_err"], places=5)
        self.assertAlmostEqual(0.04218056953351801,
                               results[inter.id]["lower_ci"], places=5)
        self.assertAlmostEqual(2.6728238164026799,
                               results[inter.id]["upper_ci"], places=5)
        self.assertAlmostEqual(2.0228173548946042,
                               results[inter.id]["z_value"], places=5)
        self.assertAlmostEqual(-np.log10(4.309198170818540e-02),
                               -np.log10(results[inter.id]["p_value"]),
                               places=5)

        # TODO: Check the other predictors

    def test_mixedlm_snp1_inter_ml(self):
        """Tests mixedlm regression with the first SNP (using ML, inter)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)
        visit = spec.factor(spec.phenotypes.visit)

        # The interaction term
        inter = spec.interaction(spec.genotypes.snp1, spec.phenotypes.var1)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome={"outcome": spec.phenotypes.pheno3,
                     "groups": spec.phenotypes.sampleid},
            predictors=[spec.genotypes.snp1, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, visit, inter],
            test=lambda: StatsMixedLM(reml=False),
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp1", results["snp1"]["name"])
        self.assertEqual("3", results["snp1"]["chrom"])
        self.assertEqual(1234, results["snp1"]["pos"])
        self.assertEqual("T", results["snp1"]["minor"])
        self.assertEqual("G", results["snp1"]["major"])
        self.assertAlmostEqual(0.17499166666666666, results["snp1"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(1.3575021929662827, results[inter.id]["coef"])
        self.assertAlmostEqual(0.6366563415656629,
                               results[inter.id]["std_err"], places=6)
        self.assertAlmostEqual(0.1096786929685527,
                               results[inter.id]["lower_ci"], places=6)
        self.assertAlmostEqual(2.6053256929640130,
                               results[inter.id]["upper_ci"], places=6)
        self.assertAlmostEqual(2.1322369767462277,
                               results[inter.id]["z_value"], places=6)
        self.assertAlmostEqual(-np.log10(3.298737010780739e-02),
                               -np.log10(results[inter.id]["p_value"]),
                               places=5)

        # TODO: Check the other predictors

    def test_mixedlm_snp2_reml(self):
        """Tests mixedlm regression with the second SNP (using REML)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)
        visit = spec.factor(spec.phenotypes.visit)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome={"outcome": spec.phenotypes.pheno3,
                     "groups": spec.phenotypes.sampleid},
            predictors=[spec.genotypes.snp2, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, visit],
            test="mixedlm",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp2", results["snp2"]["name"])
        self.assertEqual("3", results["snp2"]["chrom"])
        self.assertEqual(2345, results["snp2"]["pos"])
        self.assertEqual("C", results["snp2"]["minor"])
        self.assertEqual("A", results["snp2"]["major"])
        self.assertAlmostEqual(0.36666666666666664, results["snp2"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(28.87619649886422479, results["snp2"]["coef"])
        self.assertAlmostEqual(7.616420586507464, results["snp2"]["std_err"],
                               places=4)
        self.assertAlmostEqual(13.9482864582001636,
                               results["snp2"]["lower_ci"], places=4)
        self.assertAlmostEqual(43.8041065395282843,
                               results["snp2"]["upper_ci"], places=4)
        self.assertAlmostEqual(3.7913080259798924,
                               results["snp2"]["z_value"], places=4)
        self.assertAlmostEqual(-np.log10(1.498559584953707e-04),
                               -np.log10(results["snp2"]["p_value"]), places=4)

        # TODO: Check the other predictors

    def test_mixedlm_snp2_ml(self):
        """Tests mixedlm regression with the second SNP (using ML)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)
        visit = spec.factor(spec.phenotypes.visit)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome={"outcome": spec.phenotypes.pheno3,
                     "groups": spec.phenotypes.sampleid},
            predictors=[spec.genotypes.snp2, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, visit],
            test=lambda: StatsMixedLM(reml=False),
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(28.87619649885465023, results["snp2"]["coef"])
        self.assertAlmostEqual(7.2921675410708389, results["snp2"]["std_err"],
                               places=4)
        self.assertAlmostEqual(14.5838107491238045,
                               results["snp2"]["lower_ci"], places=4)
        self.assertAlmostEqual(43.1685822485854942,
                               results["snp2"]["upper_ci"], places=4)
        self.assertAlmostEqual(3.9598920809510423,
                               results["snp2"]["z_value"], places=4)
        self.assertAlmostEqual(-np.log10(7.498364115088307e-05),
                               -np.log10(results["snp2"]["p_value"]), places=4)

        # TODO: Check the other predictors

    def test_mixedlm_snp2_inter_reml(self):
        """Tests mixedlm regression with the second SNP (using REML, inter)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)
        visit = spec.factor(spec.phenotypes.visit)

        # The interaction term
        inter = spec.interaction(spec.genotypes.snp2, spec.phenotypes.var1)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome={"outcome": spec.phenotypes.pheno3,
                     "groups": spec.phenotypes.sampleid},
            predictors=[spec.genotypes.snp2, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, visit, inter],
            test="mixedlm",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp2", results["snp2"]["name"])
        self.assertEqual("3", results["snp2"]["chrom"])
        self.assertEqual(2345, results["snp2"]["pos"])
        self.assertEqual("C", results["snp2"]["minor"])
        self.assertEqual("A", results["snp2"]["major"])
        self.assertAlmostEqual(0.36666666666666664, results["snp2"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(0.8952450482655012, results[inter.id]["coef"])
        self.assertAlmostEqual(0.5976670951727925,
                               results[inter.id]["std_err"], places=5)
        self.assertAlmostEqual(-0.2761609330178445,
                               results[inter.id]["lower_ci"], places=5)
        self.assertAlmostEqual(2.0666510295488472,
                               results[inter.id]["upper_ci"], places=5)
        self.assertAlmostEqual(1.497899174132508504,
                               results[inter.id]["z_value"], places=5)
        self.assertAlmostEqual(-np.log10(1.341594483012889e-01),
                               -np.log10(results[inter.id]["p_value"]),
                               places=5)

        # TODO: Check the other predictors

    def test_mixedlm_snp2_inter_ml(self):
        """Tests mixedlm regression with the second SNP (using ML, inter)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)
        visit = spec.factor(spec.phenotypes.visit)

        # The interaction term
        inter = spec.interaction(spec.genotypes.snp2, spec.phenotypes.var1)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome={"outcome": spec.phenotypes.pheno3,
                     "groups": spec.phenotypes.sampleid},
            predictors=[spec.genotypes.snp2, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, visit, inter],
            test=lambda: StatsMixedLM(reml=False),
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp2", results["snp2"]["name"])
        self.assertEqual("3", results["snp2"]["chrom"])
        self.assertEqual(2345, results["snp2"]["pos"])
        self.assertEqual("C", results["snp2"]["minor"])
        self.assertEqual("A", results["snp2"]["major"])
        self.assertAlmostEqual(0.36666666666666664, results["snp2"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(0.8952450482657273, results[inter.id]["coef"])
        self.assertAlmostEqual(0.5669968515485548,
                               results[inter.id]["std_err"], places=6)
        self.assertAlmostEqual(-0.2160483601170435,
                               results[inter.id]["lower_ci"], places=5)
        self.assertAlmostEqual(2.006538456648498,
                               results[inter.id]["upper_ci"], places=5)
        self.assertAlmostEqual(1.578924196528916468,
                               results[inter.id]["z_value"], places=5)
        self.assertAlmostEqual(-np.log10(1.143534452552892e-01),
                               -np.log10(results[inter.id]["p_value"]),
                               places=5)

        # TODO: Check the other predictors

    def test_mixedlm_snp3_reml(self):
        """Tests mixedlm regression with the third SNP (using REML)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)
        visit = spec.factor(spec.phenotypes.visit)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome={"outcome": spec.phenotypes.pheno3,
                     "groups": spec.phenotypes.sampleid},
            predictors=[spec.genotypes.snp3, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, visit],
            test="mixedlm",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp3", results["snp3"]["name"])
        self.assertEqual("2", results["snp3"]["chrom"])
        self.assertEqual(3456, results["snp3"]["pos"])
        self.assertEqual("G", results["snp3"]["minor"])
        self.assertEqual("T", results["snp3"]["major"])
        self.assertAlmostEqual(0.40833333333333333, results["snp3"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(21.61350866000438486, results["snp3"]["coef"])
        self.assertAlmostEqual(6.4018199962254876, results["snp3"]["std_err"],
                               places=4)
        self.assertAlmostEqual(9.0661720318940873,
                               results["snp3"]["lower_ci"], places=4)
        self.assertAlmostEqual(34.160845288114686,
                               results["snp3"]["upper_ci"], places=4)
        self.assertAlmostEqual(3.37615063727935283,
                               results["snp3"]["z_value"], places=4)
        self.assertAlmostEqual(-np.log10(7.350766113720653e-04),
                               -np.log10(results["snp3"]["p_value"]), places=4)

        # TODO: Check the other predictors

    def test_mixedlm_snp3_ml(self):
        """Tests mixedlm regression with the third SNP (using ML)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)
        visit = spec.factor(spec.phenotypes.visit)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome={"outcome": spec.phenotypes.pheno3,
                     "groups": spec.phenotypes.sampleid},
            predictors=[spec.genotypes.snp3, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, visit],
            test=lambda: StatsMixedLM(reml=False),
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp3", results["snp3"]["name"])
        self.assertEqual("2", results["snp3"]["chrom"])
        self.assertEqual(3456, results["snp3"]["pos"])
        self.assertEqual("G", results["snp3"]["minor"])
        self.assertEqual("T", results["snp3"]["major"])
        self.assertAlmostEqual(0.40833333333333333, results["snp3"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(21.61350866001425786, results["snp3"]["coef"])
        self.assertAlmostEqual(6.1292761423280124, results["snp3"]["std_err"],
                               places=4)
        self.assertAlmostEqual(9.6003481697507578,
                               results["snp3"]["lower_ci"], places=4)
        self.assertAlmostEqual(33.6266691502777562,
                               results["snp3"]["upper_ci"], places=4)
        self.assertAlmostEqual(3.52627425459820243,
                               results["snp3"]["z_value"], places=5)
        self.assertAlmostEqual(-np.log10(4.214502895376615e-04),
                               -np.log10(results["snp3"]["p_value"]), places=4)

        # TODO: Check the other predictors

    def test_mixedlm_snp3_inter_reml(self):
        """Tests mixedlm regression with the third SNP (using REML, inter)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)
        visit = spec.factor(spec.phenotypes.visit)

        # The interaction term
        inter = spec.interaction(spec.genotypes.snp3, spec.phenotypes.var1)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome={"outcome": spec.phenotypes.pheno3,
                     "groups": spec.phenotypes.sampleid},
            predictors=[spec.genotypes.snp3, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, visit, inter],
            test="mixedlm",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp3", results["snp3"]["name"])
        self.assertEqual("2", results["snp3"]["chrom"])
        self.assertEqual(3456, results["snp3"]["pos"])
        self.assertEqual("G", results["snp3"]["minor"])
        self.assertEqual("T", results["snp3"]["major"])
        self.assertAlmostEqual(0.40833333333333333, results["snp3"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(0.9199422369515684, results[inter.id]["coef"])
        self.assertAlmostEqual(0.4498593164720226,
                               results[inter.id]["std_err"], places=5)
        self.assertAlmostEqual(0.03823417855659805,
                               results[inter.id]["lower_ci"], places=5)
        self.assertAlmostEqual(1.8016502953465388,
                               results[inter.id]["upper_ci"], places=5)
        self.assertAlmostEqual(2.0449553966473895,
                               results[inter.id]["z_value"], places=5)
        self.assertAlmostEqual(-np.log10(4.085925566922222e-02),
                               -np.log10(results[inter.id]["p_value"]),
                               places=4)

        # TODO: Check the other predictors

    def test_mixedlm_snp3_inter_ml(self):
        """Tests mixedlm regression with the third SNP (using ML, inter)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)
        visit = spec.factor(spec.phenotypes.visit)

        # The interaction term
        inter = spec.interaction(spec.genotypes.snp3, spec.phenotypes.var1)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome={"outcome": spec.phenotypes.pheno3,
                     "groups": spec.phenotypes.sampleid},
            predictors=[spec.genotypes.snp3, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, visit, inter],
            test=lambda: StatsMixedLM(reml=False),
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp3", results["snp3"]["name"])
        self.assertEqual("2", results["snp3"]["chrom"])
        self.assertEqual(3456, results["snp3"]["pos"])
        self.assertEqual("G", results["snp3"]["minor"])
        self.assertEqual("T", results["snp3"]["major"])
        self.assertAlmostEqual(0.40833333333333333, results["snp3"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(0.9199422369518273, results[inter.id]["coef"])
        self.assertAlmostEqual(0.4267740150554054,
                               results[inter.id]["std_err"], places=6)
        self.assertAlmostEqual(0.08348053790567811,
                               results[inter.id]["lower_ci"], places=5)
        self.assertAlmostEqual(1.7564039359979766,
                               results[inter.id]["upper_ci"], places=5)
        self.assertAlmostEqual(2.1555722806422435,
                               results[inter.id]["z_value"], places=5)
        self.assertAlmostEqual(-np.log10(3.111707895646454e-02),
                               -np.log10(results[inter.id]["p_value"]),
                               places=5)

        # TODO: Check the other predictors

    def test_mixedlm_snp4_reml(self):
        """Tests mixedlm regression with the fourth SNP (using REML)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)
        visit = spec.factor(spec.phenotypes.visit)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome={"outcome": spec.phenotypes.pheno3,
                     "groups": spec.phenotypes.sampleid},
            predictors=[spec.genotypes.snp4, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, visit],
            test="mixedlm",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp4", results["snp4"]["name"])
        self.assertEqual("22", results["snp4"]["chrom"])
        self.assertEqual(4567, results["snp4"]["pos"])
        self.assertEqual("A", results["snp4"]["minor"])
        self.assertEqual("C", results["snp4"]["major"])
        self.assertAlmostEqual(0.44166666666666665, results["snp4"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(1.4891822461435404, results["snp4"]["coef"])
        self.assertAlmostEqual(7.9840787818167422, results["snp4"]["std_err"],
                               places=3)
        self.assertAlmostEqual(-14.1593246159476980,
                               results["snp4"]["lower_ci"], places=3)
        self.assertAlmostEqual(17.1376891082347775,
                               results["snp4"]["upper_ci"], places=3)
        self.assertAlmostEqual(0.1865189819437983,
                               results["snp4"]["z_value"], places=5)
        self.assertAlmostEqual(-np.log10(8.520377946017625e-01),
                               -np.log10(results["snp4"]["p_value"]), places=5)

        # TODO: Check the other predictors

    def test_mixedlm_snp4_ml(self):
        """Tests mixedlm regression with the fourth SNP (using ML)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)
        visit = spec.factor(spec.phenotypes.visit)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome={"outcome": spec.phenotypes.pheno3,
                     "groups": spec.phenotypes.sampleid},
            predictors=[spec.genotypes.snp4, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, visit],
            test=lambda: StatsMixedLM(reml=False),
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp4", results["snp4"]["name"])
        self.assertEqual("22", results["snp4"]["chrom"])
        self.assertEqual(4567, results["snp4"]["pos"])
        self.assertEqual("A", results["snp4"]["minor"])
        self.assertEqual("C", results["snp4"]["major"])
        self.assertAlmostEqual(0.44166666666666665, results["snp4"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(1.4891822461364390, results["snp4"]["coef"])
        self.assertAlmostEqual(7.6441729657793820, results["snp4"]["std_err"],
                               places=4)
        self.assertAlmostEqual(-13.4931214583858772,
                               results["snp4"]["lower_ci"], places=3)
        self.assertAlmostEqual(16.4714859506587565,
                               results["snp4"]["upper_ci"], places=3)
        self.assertAlmostEqual(0.1948127355049462,
                               results["snp4"]["z_value"], places=5)
        self.assertAlmostEqual(-np.log10(8.455395518199895e-01),
                               -np.log10(results["snp4"]["p_value"]), places=6)

        # TODO: Check the other predictors

    def test_mixedlm_snp4_inter_reml(self):
        """Tests mixedlm regression with the fourth SNP (using REML, inter)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)
        visit = spec.factor(spec.phenotypes.visit)

        # The interaction term
        inter = spec.interaction(spec.genotypes.snp4, spec.phenotypes.var1)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome={"outcome": spec.phenotypes.pheno3,
                     "groups": spec.phenotypes.sampleid},
            predictors=[spec.genotypes.snp4, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, visit, inter],
            test="mixedlm",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp4", results["snp4"]["name"])
        self.assertEqual("22", results["snp4"]["chrom"])
        self.assertEqual(4567, results["snp4"]["pos"])
        self.assertEqual("A", results["snp4"]["minor"])
        self.assertEqual("C", results["snp4"]["major"])
        self.assertAlmostEqual(0.44166666666666665, results["snp4"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(0.05832811192587545, results[inter.id]["coef"])
        self.assertAlmostEqual(0.7381058671562147,
                               results[inter.id]["std_err"], places=4)
        self.assertAlmostEqual(-1.388332804478011,
                               results[inter.id]["lower_ci"], places=4)
        self.assertAlmostEqual(1.504989028329762,
                               results[inter.id]["upper_ci"], places=4)
        self.assertAlmostEqual(0.07902404590089884,
                               results[inter.id]["z_value"], places=5)
        self.assertAlmostEqual(-np.log10(9.370134970059800e-01),
                               -np.log10(results[inter.id]["p_value"]),
                               places=6)

        # TODO: Check the other predictors

    def test_mixedlm_snp4_inter_ml(self):
        """Tests mixedlm regression with the fourth SNP (using ML, inter)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)
        visit = spec.factor(spec.phenotypes.visit)

        # The interaction term
        inter = spec.interaction(spec.genotypes.snp4, spec.phenotypes.var1)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome={"outcome": spec.phenotypes.pheno3,
                     "groups": spec.phenotypes.sampleid},
            predictors=[spec.genotypes.snp4, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, visit, inter],
            test=lambda: StatsMixedLM(reml=False),
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp4", results["snp4"]["name"])
        self.assertEqual("22", results["snp4"]["chrom"])
        self.assertEqual(4567, results["snp4"]["pos"])
        self.assertEqual("A", results["snp4"]["minor"])
        self.assertEqual("C", results["snp4"]["major"])
        self.assertAlmostEqual(0.44166666666666665, results["snp4"]["maf"])

        # Checking the marker statistics (according to R lme4)
        self.assertAlmostEqual(0.05832811192492895, results[inter.id]["coef"])
        self.assertAlmostEqual(0.7002288518048638,
                               results[inter.id]["std_err"], places=5)
        self.assertAlmostEqual(-1.314095218548438,
                               results[inter.id]["lower_ci"], places=4)
        self.assertAlmostEqual(1.430751442398297,
                               results[inter.id]["upper_ci"], places=4)
        self.assertAlmostEqual(0.08329864125790624,
                               results[inter.id]["z_value"], places=6)
        self.assertAlmostEqual(-np.log10(9.336140806606035e-01),
                               -np.log10(results[inter.id]["p_value"]),
                               places=6)

        # TODO: Check the other predictors



"""
"""


# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.

__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "CC BY-NC 4.0"


from . import core
from .core import ModelSpec
from ..statistics import model_map
from .grammar import parse_modelspec, parse_formula


__all__ = [
    "ModelSpec", "parse_modelspec", "parse_formula", "result", "phenotypes",
    "genotypes", "factor", "log10", "pow", "interaction", "SNPs", "_reset",
    "pheWAS", "modelspec_from_formula",
]


result = core.result

phenotypes = core.phenotypes
genotypes = core.genotypes

factor = core.factor
log10 = core.log10
ln = core.ln
pow = core.pow
interaction = core.interaction

SNPs = core.SNPs

_reset = core._reset

PheWAS = core.PheWAS


def modelspec_from_formula(formula, test, test_kwargs):
    """Generates the model specification (and the subgroups) from formula."""
    model = parse_formula(formula)

    # Handle the statistical test.
    if test_kwargs is None:
        test_kwargs = {}

    if hasattr(test, "__call__"):
        model["test"] = lambda: test(**test_kwargs)
    else:
        model["test"] = lambda: model_map[test](**test_kwargs)

    # Handle the conditions and stratification.
    conditions = model.pop("conditions")
    if conditions is not None:
        model["stratify_by"] = [i["name"] for i in conditions]
        subgroups = [i["level"] for i in conditions]
    else:
        subgroups = None

    modelspec = ModelSpec(**model)

    return modelspec, subgroups



"""
"""


# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import statsmodels.api as sm
from statsmodels.tools.sm_exceptions import PerfectSeparationError

from ..core import StatsModels, StatsError

import numpy as np


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


__all__ = ["StatsLogistic"]


class StatsLogistic(StatsModels):
    def fit(self, y, X):
        """Fit the model.

        Args:
            y (pandas.DataFrame): The vector of endogenous variable.
            X (pandas.DataFrame): The matrix of exogenous variables.

        """
        # Creating the GLM model from StatsModels and fitting it
        model = sm.GLM(y, X, family=sm.families.Binomial())
        try:
            fitted = model.fit()

        except PerfectSeparationError as e:
            raise StatsError(str(e))

        out = {}
        parameters = fitted.params.index

        # Results about the model fit.
        out = {
            "MODEL": {
                "log_likelihood": fitted.llf,
                "nobs": fitted.nobs
            },
        }

        # Getting the confidence intervals
        conf_ints = fitted.conf_int()

        for param in parameters:
            # If GWAS, check that inference could be done on the SNP.
            if param == "SNPs" and np.isnan(fitted.pvalues[param]):
                raise StatsError("Inference did not converge.")

            out[param] = {
                "coef": fitted.params[param],
                "std_err": fitted.bse[param],
                "lower_ci": conf_ints.loc[param, 0],
                "upper_ci": conf_ints.loc[param, 1],
                "t_value": fitted.tvalues[param],
                "p_value": fitted.pvalues[param],
            }

        return out





# This file is part of genetest.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial
# 4.0 International License. To view a copy of this license, visit
# http://creativecommons.org/licenses/by-nc/4.0/ or send a letter to Creative
# Commons, PO Box 1866, Mountain View, CA 94042, USA.


import os
import unittest
from tempfile import TemporaryDirectory

import numpy as np
import pandas as pd
from pkg_resources import resource_filename

from geneparse.dataframe import DataFrameReader

from ..statistics.core import StatsError

from .. import analysis
from .. import subscribers
from .. import modelspec as spec
from ..phenotypes.dummy import _DummyPhenotypes


__copyright__ = "Copyright 2016, Beaulieu-Saucier Pharmacogenomics Centre"
__license__ = "Attribution-NonCommercial 4.0 International (CC BY-NC 4.0)"


class TestStatsLogistic(unittest.TestCase):
    """Tests the 'StatsLogistic' class."""
    @classmethod
    def setUpClass(cls):
        # Loading the data
        cls.data = pd.read_csv(
            resource_filename(__name__, "data/statistics/logistic.txt.bz2"),
            sep="\t",
            compression="bz2",
        )

        # Creating the index
        cls.data["sample"] = [
            "s{}".format(i+1) for i in range(cls.data.shape[0])
        ]
        cls.data = cls.data.set_index("sample")

        # Creating the dummy phenotype container
        cls.phenotypes = _DummyPhenotypes()
        cls.phenotypes.data = cls.data.drop(
            ["snp{}".format(i+1) for i in range(3)],
            axis=1,
        )

        # Permuting the sample to add a bit of randomness
        new_sample_order = np.random.permutation(cls.data.index)

        # Creating the genotypes data frame
        genotypes = cls.data.loc[
            new_sample_order,
            [_ for _ in cls.data.columns if _.startswith("snp")],
        ].copy()

        # Creating the mapping information
        map_info = pd.DataFrame(
            {"chrom": ["3", "3", "2"],
             "pos": [1234, 9618, 1519],
             "a1": ["T", "C", "G"],
             "a2": ["C", "A", "T"]},
            index=["snp1", "snp2", "snp3"],
        )

        # Creating the genotype parser
        cls.genotypes = DataFrameReader(
            dataframe=genotypes,
            map_info=map_info,
        )

        # Creating a temporary directory
        cls.tmp_dir = TemporaryDirectory(prefix="genetest_test_logistic_")

    def setUp(self):
        # Resetting the model specification
        spec._reset()

        # Reordering the columns and the rows of the phenotype data frame
        self.phenotypes.data = self.phenotypes.data.iloc[
            np.random.permutation(self.phenotypes.data.shape[0]),
            np.random.permutation(self.phenotypes.data.shape[1])
        ]

    @classmethod
    def tearDownClass(cls):
        # Cleaning the temporary directory
        cls.tmp_dir.cleanup()

    def test_logistic_gwas(self):
        """Tests logistic regression for GWAS."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno2,
            predictors=[spec.SNPs, spec.phenotypes.age,
                        spec.phenotypes.var1, gender],
            test="logistic",
        )

        # The output prefix
        out_prefix = os.path.join(self.tmp_dir.name, "results")

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber], output_prefix=out_prefix,
        )
        gwas_results = subscriber._get_gwas_results()

        # Checking the number of results (should be 2)
        self.assertEqual(2, len(gwas_results.keys()))

        # Checking the first marker (snp1)
        results = gwas_results["snp1"]
        self.assertEqual("snp1", results["SNPs"]["name"])
        self.assertEqual("3", results["SNPs"]["chrom"])
        self.assertEqual(1234, results["SNPs"]["pos"])
        self.assertEqual("T", results["SNPs"]["minor"])
        self.assertEqual("C", results["SNPs"]["major"])
        self.assertAlmostEqual(0.3749333333333334, results["SNPs"]["maf"])

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(-2.24198635855498, results["SNPs"]["coef"],
                               places=5)
        self.assertAlmostEqual(0.59759558908668, results["SNPs"]["std_err"],
                               places=6)
        self.assertAlmostEqual(-3.41325219048488, results["SNPs"]["lower_ci"],
                               places=5)
        self.assertAlmostEqual(-1.07072052662507, results["SNPs"]["upper_ci"],
                               places=5)
        self.assertAlmostEqual(14.0750894991982, results["SNPs"]["t_value"]**2,
                               places=4)
        self.assertAlmostEqual(-np.log10(0.0001756548178104),
                               -np.log10(results["SNPs"]["p_value"]),
                               places=5)

        # Checking the second marker (snp2)
        results = gwas_results["snp2"]
        self.assertEqual("snp2", results["SNPs"]["name"])
        self.assertEqual("3", results["SNPs"]["chrom"])
        self.assertEqual(9618, results["SNPs"]["pos"])
        self.assertEqual("C", results["SNPs"]["minor"])
        self.assertEqual("A", results["SNPs"]["major"])
        self.assertAlmostEqual(0.41590833333333332, results["SNPs"]["maf"])

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(1.12532308347075, results["SNPs"]["coef"],
                               places=5)
        self.assertAlmostEqual(0.45211815097241, results["SNPs"]["std_err"],
                               places=6)
        self.assertAlmostEqual(0.23918779080797, results["SNPs"]["lower_ci"],
                               places=5)
        self.assertAlmostEqual(2.01145837613353, results["SNPs"]["upper_ci"],
                               places=5)
        self.assertAlmostEqual(6.19513207316499, results["SNPs"]["t_value"]**2,
                               places=4)
        self.assertAlmostEqual(-np.log10(0.0128102164253392),
                               -np.log10(results["SNPs"]["p_value"]),
                               places=4)

        # There should be a file for the failed snp3
        self.assertTrue(os.path.isfile(out_prefix + "_failed_snps.txt"))
        with open(out_prefix + "_failed_snps.txt") as f:
            self.assertEqual(
                [["snp3", "Perfect separation detected, results not "
                          "available"]],
                [line.split("\t") for line in f.read().splitlines()],
            )

    def test_logistic_snp1(self):
        """Tests logistic regression with the first SNP."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno2,
            predictors=[spec.genotypes.snp1, spec.phenotypes.age,
                        spec.phenotypes.var1, gender],
            test="logistic",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp1", results["snp1"]["name"])
        self.assertEqual("3", results["snp1"]["chrom"])
        self.assertEqual(1234, results["snp1"]["pos"])
        self.assertEqual("T", results["snp1"]["minor"])
        self.assertEqual("C", results["snp1"]["major"])
        self.assertAlmostEqual(0.3749333333333334, results["snp1"]["maf"])

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(-2.24198635855498, results["snp1"]["coef"],
                               places=5)
        self.assertAlmostEqual(0.59759558908668, results["snp1"]["std_err"],
                               places=6)
        self.assertAlmostEqual(-3.41325219048488, results["snp1"]["lower_ci"],
                               places=5)
        self.assertAlmostEqual(-1.07072052662507, results["snp1"]["upper_ci"],
                               places=5)
        self.assertAlmostEqual(14.0750894991982, results["snp1"]["t_value"]**2,
                               places=4)
        self.assertAlmostEqual(-np.log10(0.0001756548178104),
                               -np.log10(results["snp1"]["p_value"]),
                               places=5)

        # TODO: Check the other predictors

    def test_logistic_snp1_inter(self):
        """Tests logistic regression with the first SNP (interaction)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # The interaction term
        inter = spec.interaction(spec.phenotypes.var1, spec.genotypes.snp1)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno2,
            predictors=[spec.genotypes.snp1, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, inter],
            test="logistic",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp1", results["snp1"]["name"])
        self.assertEqual("3", results["snp1"]["chrom"])
        self.assertEqual(1234, results["snp1"]["pos"])
        self.assertEqual("T", results["snp1"]["minor"])
        self.assertEqual("C", results["snp1"]["major"])
        self.assertAlmostEqual(0.3749333333333334, results["snp1"]["maf"])

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(0.0015311482968189, results[inter.id]["coef"],
                               places=6)
        self.assertAlmostEqual(0.0423504567007674,
                               results[inter.id]["std_err"])
        self.assertAlmostEqual(-0.0814742215655, results[inter.id]["lower_ci"],
                               places=6)
        self.assertAlmostEqual(0.08453651815914, results[inter.id]["upper_ci"])
        self.assertAlmostEqual(0.0013071285938733,
                               results[inter.id]["t_value"]**2, places=6)
        self.assertAlmostEqual(0.9711593785536400,
                               results[inter.id]["p_value"], places=5)

        # TODO: Check the other predictors

    def test_logistic_snp1_inter_categorical(self):
        """Tests logistic regression first SNP (inter, category)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # The interaction term
        inter = spec.interaction(gender, spec.genotypes.snp1)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno2,
            predictors=[spec.genotypes.snp1, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, inter],
            test="logistic",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp1", results["snp1"]["name"])
        self.assertEqual("3", results["snp1"]["chrom"])
        self.assertEqual(1234, results["snp1"]["pos"])
        self.assertEqual("T", results["snp1"]["minor"])
        self.assertEqual("C", results["snp1"]["major"])
        self.assertAlmostEqual(0.3749333333333334, results["snp1"]["maf"])

        # Checking the marker statistics (according to SAS)
        col = inter.id + ":level.2"
        self.assertAlmostEqual(-0.0089544870073693, results[col]["coef"],
                               places=4)
        self.assertAlmostEqual(1.19878325025509, results[col]["std_err"],
                               places=5)
        self.assertAlmostEqual(-2.35852648277723, results[col]["lower_ci"],
                               places=4)
        self.assertAlmostEqual(2.34061750876249, results[col]["upper_ci"],
                               places=5)
        self.assertAlmostEqual(np.log10(0.0000557956175619),
                               np.log10(results[col]["t_value"]**2), places=2)
        self.assertAlmostEqual(0.99404013987338, results[col]["p_value"],
                               places=5)

        # TODO: Check the other predictors

    def test_logistic_snp2(self):
        """Tests logistic regression with the second SNP."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno2,
            predictors=[spec.genotypes.snp2, spec.phenotypes.age,
                        spec.phenotypes.var1, gender],
            test="logistic",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp2", results["snp2"]["name"])
        self.assertEqual("3", results["snp2"]["chrom"])
        self.assertEqual(9618, results["snp2"]["pos"])
        self.assertEqual("C", results["snp2"]["minor"])
        self.assertEqual("A", results["snp2"]["major"])
        self.assertAlmostEqual(0.41590833333333332, results["snp2"]["maf"])

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(1.12532308347075, results["snp2"]["coef"],
                               places=5)
        self.assertAlmostEqual(0.45211815097241, results["snp2"]["std_err"],
                               places=6)
        self.assertAlmostEqual(0.23918779080797, results["snp2"]["lower_ci"],
                               places=5)
        self.assertAlmostEqual(2.01145837613353, results["snp2"]["upper_ci"],
                               places=5)
        self.assertAlmostEqual(6.19513207316499, results["snp2"]["t_value"]**2,
                               places=4)
        self.assertAlmostEqual(-np.log10(0.0128102164253392),
                               -np.log10(results["snp2"]["p_value"]),
                               places=4)

        # TODO: Check the other predictors

    def test_logistic_snp2_inter(self):
        """Tests logistic regression with the second SNP (interaction)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # The interaction term
        inter = spec.interaction(spec.phenotypes.var1, spec.genotypes.snp2)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno2,
            predictors=[spec.genotypes.snp2, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, inter],
            test="logistic",
        )

        # Performing the analysis and retrieving the results
        subscriber = subscribers.ResultsMemory()
        analysis.execute(
            self.phenotypes, self.genotypes, modelspec,
            subscribers=[subscriber],
        )
        results = subscriber.results[0]

        # Checking the marker information
        self.assertEqual("snp2", results["snp2"]["name"])
        self.assertEqual("3", results["snp2"]["chrom"])
        self.assertEqual(9618, results["snp2"]["pos"])
        self.assertEqual("C", results["snp2"]["minor"])
        self.assertEqual("A", results["snp2"]["major"])
        self.assertAlmostEqual(0.41590833333333332, results["snp2"]["maf"])

        # Checking the marker statistics (according to SAS)
        self.assertAlmostEqual(0.0239292800721822, results[inter.id]["coef"],
                               places=5)
        self.assertAlmostEqual(0.0342619407842591,
                               results[inter.id]["std_err"], places=6)
        self.assertAlmostEqual(-0.0432228899054096,
                               results[inter.id]["lower_ci"], places=5)
        self.assertAlmostEqual(0.09108145004977, results[inter.id]["upper_ci"],
                               places=5)
        self.assertAlmostEqual(0.48779275460699,
                               results[inter.id]["t_value"]**2, places=3)
        self.assertAlmostEqual(0.48491356159603, results[inter.id]["p_value"],
                               places=4)

        # TODO: Check the other predictors

    def test_logistic_snp3(self):
        """Tests logistic regression with the third SNP (raises StatsError)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno2,
            predictors=[spec.genotypes.snp3, spec.phenotypes.age,
                        spec.phenotypes.var1, gender],
            test="logistic",
        )

        # Performing the analysis and retrieving the results
        with self.assertRaises(StatsError) as cm:
            analysis.execute(
                self.phenotypes, self.genotypes, modelspec,
                subscribers=[subscribers.ResultsMemory()],
            )
        self.assertEqual("Perfect separation detected, results not available",
                         str(cm.exception))

    def test_logistic_snp3_inter(self):
        """Tests logistic regression third SNP (raises StatsError, inter)."""
        # The variables which are factors
        gender = spec.factor(spec.phenotypes.gender)

        # The interaction term
        inter = spec.interaction(spec.phenotypes.var1, spec.genotypes.snp2)

        # Creating the model specification
        modelspec = spec.ModelSpec(
            outcome=spec.phenotypes.pheno2,
            predictors=[spec.genotypes.snp3, spec.phenotypes.age,
                        spec.phenotypes.var1, gender, inter],
            test="logistic",
        )

        # Performing the analysis and retrieving the results
        with self.assertRaises(StatsError) as cm:
            analysis.execute(
                self.phenotypes, self.genotypes, modelspec,
                subscribers=[subscribers.ResultsMemory()],
            )
        self.assertEqual("Perfect separation detected, results not available",
                         str(cm.exception))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# CAVEAT UTILITOR
#
# This file was automatically generated by Grako.
#
#    https://pypi.python.org/pypi/grako/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.


from __future__ import print_function, division, absolute_import, unicode_literals

from grako.buffering import Buffer
from grako.parsing import graken, Parser
from grako.util import re, RE_FLAGS, generic_main  # noqa


KEYWORDS = {}


class ModelSpecBuffer(Buffer):
    def __init__(
        self,
        text,
        whitespace=None,
        nameguard=None,
        comments_re=None,
        eol_comments_re=None,
        ignorecase=None,
        namechars='',
        **kwargs
    ):
        super(ModelSpecBuffer, self).__init__(
            text,
            whitespace=whitespace,
            nameguard=nameguard,
            comments_re=comments_re,
            eol_comments_re=eol_comments_re,
            ignorecase=ignorecase,
            namechars=namechars,
            **kwargs
        )


class ModelSpecParser(Parser):
    def __init__(
        self,
        whitespace=None,
        nameguard=None,
        comments_re=None,
        eol_comments_re=None,
        ignorecase=None,
        left_recursion=False,
        parseinfo=True,
        keywords=None,
        namechars='',
        buffer_class=ModelSpecBuffer,
        **kwargs
    ):
        if keywords is None:
            keywords = KEYWORDS
        super(ModelSpecParser, self).__init__(
            whitespace=whitespace,
            nameguard=nameguard,
            comments_re=comments_re,
            eol_comments_re=eol_comments_re,
            ignorecase=ignorecase,
            left_recursion=left_recursion,
            parseinfo=parseinfo,
            keywords=keywords,
            namechars=namechars,
            buffer_class=buffer_class,
            **kwargs
        )

    @graken()
    def _model_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._phenotype_or_variant_()
                with self._option():
                    self._labelled_outcome_group_()
                self._error('no available options')
        self.name_last_node('outcome')
        with self._optional():
            self._token('|')
            self._cut()
            self._condition_group_()
            self.name_last_node('conditions')
        self._token('~')
        self._predictors_()
        self.name_last_node('predictors')
        self._check_eof()
        self.ast._define(
            ['conditions', 'outcome', 'predictors'],
            []
        )

    @graken()
    def _condition_group_(self):

        def sep0():
            self._token(',')

        def block0():
            self._condition_()
        self._gather(block0, sep0)

    @graken()
    def _condition_(self):
        self._phenotype_or_variant_()
        self.name_last_node('name')
        with self._optional():
            self._token('=')
            self._integer_()
            self.name_last_node('level')
        self.ast._define(
            ['level', 'name'],
            []
        )

    @graken()
    def _labelled_outcome_group_(self):
        self._token('[')
        self._cut()

        def sep1():
            self._token(',')

        def block1():
            self._labelled_outcome_()
        self._gather(block1, sep1)
        self.name_last_node('tags')
        self._token(']')
        self.ast._define(
            ['tags'],
            []
        )

    @graken()
    def _labelled_outcome_(self):
        self._name_()
        self.name_last_node('key')
        self._token('=')
        self._phenotype_()
        self.name_last_node('name')
        self.ast._define(
            ['key', 'name'],
            []
        )

    @graken()
    def _genotype_(self):
        self._token('g(')
        self._cut()
        self._name_()
        self.name_last_node('variant')
        self._token(')')
        self.ast._define(
            ['variant'],
            []
        )

    @graken()
    def _factor_(self):
        self._token('factor(')
        self._cut()
        self._phenotype_()
        self.name_last_node('phen')
        self._token(')')
        with self._optional():
            self._token('as')
            self._name_()
            self.name_last_node('as_')
        self.ast._define(
            ['as_', 'phen'],
            []
        )

    @graken()
    def _pow_(self):
        self._token('pow(')
        self._cut()
        self._phenotype_()
        self.name_last_node('phen')
        self._token(', ')
        self._cut()
        self._integer_()
        self.name_last_node('power')
        self._token(')')
        with self._optional():
            self._token('as')
            self._name_()
            self.name_last_node('as_')
        self.ast._define(
            ['as_', 'phen', 'power'],
            []
        )

    @graken()
    def _ln_(self):
        self._token('ln(')
        self._cut()
        self._phenotype_()
        self.name_last_node('phen')
        self._token(')')
        with self._optional():
            self._token('as')
            self._name_()
            self.name_last_node('as_')
        self.ast._define(
            ['as_', 'phen'],
            []
        )

    @graken()
    def _log10_(self):
        self._token('log10(')
        self._cut()
        self._phenotype_()
        self.name_last_node('phen')
        self._token(')')
        with self._optional():
            self._token('as')
            self._name_()
            self.name_last_node('as_')
        self.ast._define(
            ['as_', 'phen'],
            []
        )

    @graken()
    def _predictors_(self):

        def sep0():
            self._token('+')

        def block0():
            self._expression_()
        self._positive_gather(block0, sep0)

    @graken()
    def _interaction_(self):
        with self._if():
            with self._group():
                with self._group():
                    with self._choice():
                        with self._option():
                            self._genotype_()
                        with self._option():
                            self._token('factor(')
                            self._name_()
                            self._token(')')
                        with self._option():
                            self._name_()
                        self._error('no available options')
                self._token('*')

        def sep2():
            self._token('*')

        def block2():
            with self._choice():
                with self._option():
                    self._factor_()
                with self._option():
                    self._phenotype_or_variant_()
                self._error('no available options')
        self._gather(block2, sep2)
        self.name_last_node('interaction')
        with self._optional():
            self._token('as')
            self._name_()
            self.name_last_node('as_')
        self.ast._define(
            ['as_', 'interaction'],
            []
        )

    @graken()
    def _SNPs_(self):
        self._token('SNPs')

    @graken()
    def _phenotype_(self):
        self._name_()
        self.name_last_node('name')
        self.ast._define(
            ['name'],
            []
        )

    @graken()
    def _name_(self):
        self._pattern(r'[A-Za-z0-9_:]+')

    @graken()
    def _phenotype_or_variant_(self):
        with self._choice():
            with self._option():
                self._genotype_()
            with self._option():
                self._phenotype_()
            self._error('no available options')

    @graken()
    def _integer_(self):
        self._pattern(r'[0-9]+')
        self.name_last_node('int')
        self.ast._define(
            ['int'],
            []
        )

    @graken()
    def _expression_(self):
        with self._choice():
            with self._option():
                self._SNPs_()
            with self._option():
                self._interaction_()
            with self._option():
                self._genotype_()
            with self._option():
                self._factor_()
            with self._option():
                self._ln_()
            with self._option():
                self._log10_()
            with self._option():
                self._pow_()
            with self._option():
                self._phenotype_()
            self._error('no available options')


class ModelSpecSemantics(object):
    def model(self, ast):
        return ast

    def condition_group(self, ast):
        return ast

    def condition(self, ast):
        return ast

    def labelled_outcome_group(self, ast):
        return ast

    def labelled_outcome(self, ast):
        return ast

    def genotype(self, ast):
        return ast

    def factor(self, ast):
        return ast

    def pow(self, ast):
        return ast

    def ln(self, ast):
        return ast

    def log10(self, ast):
        return ast

    def predictors(self, ast):
        return ast

    def interaction(self, ast):
        return ast

    def SNPs(self, ast):
        return ast

    def phenotype(self, ast):
        return ast

    def name(self, ast):
        return ast

    def phenotype_or_variant(self, ast):
        return ast

    def integer(self, ast):
        return ast

    def expression(self, ast):
        return ast


def main(filename, startrule, **kwargs):
    with open(filename) as f:
        text = f.read()
    parser = ModelSpecParser()
    return parser.parse(text, startrule, filename=filename, **kwargs)


if __name__ == '__main__':
    import json
    from grako.util import asjson

    ast = generic_main(main, ModelSpecParser, name='ModelSpec')
    print('AST:')
    print(ast)
    print()
    print('JSON:')
    print(json.dumps(asjson(ast), indent=2))
    print()
